@@ -33,6 +33,8 @@
 
 #include "apputils.h"
 
+#include <event2/event.h>
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
@@ -43,10 +45,9 @@
 #include <getopt.h>
 #include <locale.h>
 #include <libgen.h>
+#include <fcntl.h>
 
-#if !defined(TURN_NO_THREADS)
 #include <pthread.h>
-#endif
 
 #include <signal.h>
 
@@ -55,8 +56,26 @@
 #include <sys/stat.h>
 #include <sys/resource.h>
 
+/************************/
+
+int IS_TURN_SERVER = 0;
+
 /*********************** Sockets *********************************/
 
+int socket_set_nonblocking(evutil_socket_t fd)
+{
+#if defined(WIN32)
+	unsigned long nonblocking = 1;
+    ioctlsocket(fd, FIONBIO, (unsigned long*) &nonblocking);
+#else
+    if (fcntl(fd, F_SETFL, O_NONBLOCK) == -1) {
+        perror("O_NONBLOCK");
+        return -1;
+    }
+#endif
+    return 0;
+}
+
 void read_spare_buffer(evutil_socket_t fd)
 {
 	if(fd >= 0) {
@@ -65,14 +84,39 @@ void read_spare_buffer(evutil_socket_t fd)
 	}
 }
 
-int set_sock_buf_size(evutil_socket_t fd, int sz) {
+int set_sock_buf_size(evutil_socket_t fd, int sz0)
+{
+	int sz;
 
-  if(setsockopt(fd,SOL_SOCKET,SO_RCVBUF,(const void*)(&sz),(socklen_t)sizeof(sz))<0) {
-    perror("Cannot set socket size");
-    TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,"Cannot set sock size on %d\n",fd);
-    return -1;
-  }
-  return 0;
+	sz = sz0;
+	while (sz > 0) {
+		if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (const void*) (&sz), (socklen_t) sizeof(sz)) < 0) {
+			sz = sz / 2;
+		} else {
+			break;
+		}
+	}
+
+	if (sz < 1) {
+		perror("Cannot set socket rcv size");
+		TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "Cannot set rcv sock size %d on fd %d\n", sz0, fd);
+	}
+
+	sz = sz0;
+	while (sz > 0) {
+		if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (const void*) (&sz), (socklen_t) sizeof(sz)) < 0) {
+			sz = sz / 2;
+		} else {
+			break;
+		}
+	}
+
+	if (sz < 1) {
+		perror("Cannot set socket snd size");
+		TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "Cannot set snd sock size %d on fd %d\n", sz0, fd);
+	}
+
+	return 0;
 }
 
 int socket_tcp_set_keepalive(evutil_socket_t fd)
@@ -80,7 +124,7 @@ int socket_tcp_set_keepalive(evutil_socket_t fd)
 #ifdef SO_KEEPALIVE
     /* Set the keepalive option active */
     {
-	    const int on = 1;
+	    int on = 1;
 	    setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (const void*)&on, (socklen_t) sizeof(on));
     }
 #else
@@ -89,7 +133,7 @@ int socket_tcp_set_keepalive(evutil_socket_t fd)
 
 #ifdef SO_NOSIGPIPE
     {
-    	 const int on = 1;
+    	 int on = 1;
     	 setsockopt(fd, SOL_SOCKET, SO_NOSIGPIPE, (const void*)&on, (socklen_t) sizeof(on));
     }
 #endif
@@ -97,37 +141,43 @@ int socket_tcp_set_keepalive(evutil_socket_t fd)
     return 0;
 }
 
-int socket_set_reusable(evutil_socket_t fd) {
-  if(fd<0) return -1;
-  else {
-    evutil_make_listen_socket_reuseable(fd);
+int socket_set_reusable(evutil_socket_t fd, int flag)
+{
 
-#ifdef SO_REUSEPORT
-    {
-      const int on = 1;
-      setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, (const void*) &on, (socklen_t) sizeof(on));
-    }
+	if (fd < 0)
+		return -1;
+	else {
+
+#if defined(WIN32)
+		int use_reuseaddr = IS_TURN_SERVER;
+#else
+		int use_reuseaddr = 1;
 #endif
 
-#ifdef SO_REUSEADDR
-    {
-      const int on = 1;
-      int ret = setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (const void*) &on, (socklen_t) sizeof(on));
-      if(ret<0)
-	perror("SO_REUSEADDR");
-    }
+#if defined(SO_REUSEPORT)
+		if (use_reuseaddr) {
+			int on = flag;
+			setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, (const void*) &on, (socklen_t) sizeof(on));
+		}
 #endif
 
-    return 0;
-  }
+#if defined(SO_REUSEADDR)
+		if (use_reuseaddr) {
+			int on = flag;
+			int ret = setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (const void*) &on, (socklen_t) sizeof(on));
+			if (ret < 0)
+				perror("SO_REUSEADDR");
+		}
+#endif
+
+		return 0;
+	}
 }
 
 int sock_bind_to_device(evutil_socket_t fd, const unsigned char* ifname) {
 
 	if (fd >= 0 && ifname && ifname[0]) {
 
-		socket_set_reusable(fd);
-
 #if defined(SO_BINDTODEVICE)
 
 		struct ifreq ifr;
@@ -160,9 +210,9 @@ int addr_connect(evutil_socket_t fd, const ioa_addr* addr, int *out_errno)
 	else {
 		int err = 0;
 		do {
-			if (addr->ss.ss_family == AF_INET) {
+			if (addr->ss.sa_family == AF_INET) {
 				err = connect(fd, (const struct sockaddr *) addr, sizeof(struct sockaddr_in));
-			} else if (addr->ss.ss_family == AF_INET6) {
+			} else if (addr->ss.sa_family == AF_INET6) {
 				err = connect(fd, (const struct sockaddr *) addr, sizeof(struct sockaddr_in6));
 			} else {
 				return -1;
@@ -179,20 +229,23 @@ int addr_connect(evutil_socket_t fd, const ioa_addr* addr, int *out_errno)
 	}
 }
 
-int addr_bind(evutil_socket_t fd, const ioa_addr* addr)
+int addr_bind(evutil_socket_t fd, const ioa_addr* addr, int reusable)
 {
-	if (!addr || fd < 0)
+	if (!addr || fd < 0) {
+
 		return -1;
-	else {
+
+	} else {
+
 		int ret = -1;
 
-		socket_set_reusable(fd);
+		socket_set_reusable(fd, reusable);
 
-		if (addr->ss.ss_family == AF_INET) {
+		if (addr->ss.sa_family == AF_INET) {
 			do {
 				ret = bind(fd, (const struct sockaddr *) addr, sizeof(struct sockaddr_in));
 			} while (ret < 0 && errno == EINTR);
-		} else if (addr->ss.ss_family == AF_INET6) {
+		} else if (addr->ss.sa_family == AF_INET6) {
 			const int off = 0;
 			setsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY, (const char *) &off, sizeof(off));
 			do {
@@ -201,6 +254,13 @@ int addr_bind(evutil_socket_t fd, const ioa_addr* addr)
 		} else {
 			return -1;
 		}
+		if(ret<0) {
+			int err = errno;
+			perror("bind");
+			char str[129];
+			addr_to_string(addr,(u08bits*)str);
+			TURN_LOG_FUNC(TURN_LOG_LEVEL_WARNING, "Trying to bind fd %d to <%s>: errno=%d\n", fd, str, err);
+		}
 		return ret;
 	}
 }
@@ -213,10 +273,10 @@ int addr_get_from_sock(evutil_socket_t fd, ioa_addr *addr)
 	else {
 
 		ioa_addr a;
-		a.ss.ss_family = AF_INET6;
+		a.ss.sa_family = AF_INET6;
 		socklen_t socklen = get_ioa_addr_len(&a);
 		if (getsockname(fd, (struct sockaddr*) &a, &socklen) < 0) {
-			a.ss.ss_family = AF_INET;
+			a.ss.sa_family = AF_INET;
 			socklen = get_ioa_addr_len(&a);
 			if (getsockname(fd, (struct sockaddr*) &a, &socklen) < 0) {
 				return -1;
@@ -389,11 +449,6 @@ int get_socket_mtu(evutil_socket_t fd, int family, int verbose)
 #endif
 		;
 	}
-	if(ret<0) {
-		TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,"%s: get IP_MTU sockopt failed\n",__FUNCTION__);
-	} else {
-		if(verbose) TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,"%s: IP_MTU=%d\n",__FUNCTION__,val);
-	}
 
 	ret = val;
 #endif
@@ -440,11 +495,8 @@ int handle_socket_error() {
     return 1;
   case ECONNRESET:
   case ECONNREFUSED:
-    /* Connection reset by peer.
-     * Just ignore, might be an attacker
-     * sending fake ICMP messages.
-     */
-    return 1;
+    /* Connection reset by peer. */
+    return 0;
   case ENOMEM:
     /* Out of memory.
      * Must close connection.
@@ -483,7 +535,13 @@ char *skip_blanks(char* s)
 #define ETCDIR INSTALL_PREFIX/etc/
 #define QETCDIR QUOTE(ETCDIR)
 
-static const char *config_file_search_dirs[] = {"./", "./etc/", "../etc/", "/etc/", "/usr/local/etc/", QETCDIR, NULL };
+#define ETCDIR1 INSTALL_PREFIX/etc/turnserver/
+#define QETCDIR1 QUOTE(ETCDIR1)
+
+#define ETCDIR2 INSTALL_PREFIX/etc/rfc5766-turn-server/
+#define QETCDIR2 QUOTE(ETCDIR2)
+
+static const char *config_file_search_dirs[] = {"./", "./turnserver/", "./rfc5766-turn-server/", "./etc/", "./etc/turnserver/", "./etc/rfc5766-turn-server/", "../etc/", "../etc/turnserver/", "../etc/rfc5766-turn-server/", "/etc/", "/etc/turnserver/", "/etc/rfc5766-turn-server/", "/usr/local/etc/", "/usr/local/etc/turnserver/", "/usr/local/etc/rfc5766-turn-server/", QETCDIR, QETCDIR1, QETCDIR2, NULL };
 static char *c_execdir=NULL;
 
 void set_execdir(void)
@@ -500,9 +558,9 @@ void set_execdir(void)
     else
       edir = dirname(_var);
     if(c_execdir)
-      free(c_execdir);
+      turn_free(c_execdir,strlen(c_execdir)+1);
     c_execdir = strdup(edir);
-    free(_var);
+    turn_free(_var,strlen(_var)+1);
   }
 }
 
@@ -542,7 +600,7 @@ char* find_config_file(const char *config_file, int print_file_name)
 	char *full_path_to_config_file = NULL;
 
 	if (config_file && config_file[0]) {
-		if (config_file[0] == '/') {
+		if ((config_file[0] == '/')||(config_file[0] == '~')) {
 			FILE *f = fopen(config_file, "r");
 			if (f) {
 				fclose(f);
@@ -555,7 +613,7 @@ char* find_config_file(const char *config_file, int print_file_name)
 			while (config_file_search_dirs[i]) {
 				size_t dirlen = strlen(config_file_search_dirs[i]);
 				size_t fnsz = sizeof(char) * (dirlen + cflen + 10);
-				char *fn = (char*)malloc(fnsz+1);
+				char *fn = (char*)turn_malloc(fnsz+1);
 				strncpy(fn, config_file_search_dirs[i], fnsz);
 				strncpy(fn + dirlen, config_file, fnsz-dirlen);
 				fn[fnsz]=0;
@@ -567,13 +625,13 @@ char* find_config_file(const char *config_file, int print_file_name)
 					full_path_to_config_file = fn;
 					break;
 				}
-				free(fn);
+				turn_free(fn,fnsz+1);
 				if(config_file_search_dirs[i][0]!='/' && 
 				   config_file_search_dirs[i][0]!='.' &&
 				   c_execdir && c_execdir[0]) {
 					size_t celen = strlen(c_execdir);
 					fnsz = sizeof(char) * (dirlen + cflen + celen + 10);
-					fn = (char*)malloc(fnsz+1);
+					fn = (char*)turn_malloc(fnsz+1);
 					strncpy(fn,c_execdir,fnsz);
 					size_t fnlen=strlen(fn);
 					if(fnlen<fnsz) {
@@ -598,7 +656,7 @@ char* find_config_file(const char *config_file, int print_file_name)
 					    break;
 					  }
 					}
-					free(fn);
+					turn_free(fn,fnsz+1);
 				}
 				++i;
 			}
@@ -616,15 +674,35 @@ char* find_config_file(const char *config_file, int print_file_name)
 
 /////////////////// SYS SETTINGS ///////////////////////
 
-void set_system_parameters(int max_resources)
+void ignore_sigpipe(void)
 {
-	srandom((unsigned int) time(NULL));
+	/* Ignore SIGPIPE from TCP sockets */
+	if(signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
+		perror("Cannot set SIGPIPE handler");
+	}
+}
+
+static u64bits turn_getRandTime(void) {
+  struct timespec tp={0,0};
+#if defined(CLOCK_REALTIME)
+  clock_gettime(CLOCK_REALTIME, &tp);
+#else
+  tp.tv_sec = time(NULL);
+#endif
+  u64bits current_time = (u64bits)(tp.tv_sec);
+  u64bits current_mstime = (u64bits)(current_time + (tp.tv_nsec));
+
+  return current_mstime;
+}
+
+unsigned long set_system_parameters(int max_resources)
+{
+	srandom((unsigned int) (turn_getRandTime() + (unsigned int)((long)(&turn_getRandTime))));
 	setlocale(LC_ALL, "C");
 
 	build_base64_decoding_table();
 
-	/* Ignore SIGPIPE from TCP sockets */
-	signal(SIGPIPE, SIG_IGN);
+	ignore_sigpipe();
 
 	if(max_resources) {
 		struct rlimit rlim;
@@ -635,8 +713,11 @@ void set_system_parameters(int max_resources)
 			while((setrlimit(RLIMIT_NOFILE, &rlim)<0) && (rlim.rlim_cur>0)) {
 				rlim.rlim_cur = rlim.rlim_cur>>1;
 			}
+			return (unsigned long)rlim.rlim_cur;
 		}
 	}
+
+	return 0;
 }
 
 ////////////////////// Base 64 ////////////////////////////
@@ -658,7 +739,7 @@ char *base64_encode(const unsigned char *data,
 
     *output_length = 4 * ((input_length + 2) / 3);
 
-    char *encoded_data = (char*)malloc(*output_length+1);
+    char *encoded_data = (char*)turn_malloc(*output_length+1);
     if (encoded_data == NULL) return NULL;
 
     size_t i,j;
@@ -686,10 +767,11 @@ char *base64_encode(const unsigned char *data,
 
 void build_base64_decoding_table() {
 
-    decoding_table = (char*)malloc(256);
+    decoding_table = (char*)turn_malloc(256);
+    ns_bzero(decoding_table,256);
 
     int i;
-    for (i = 0; i < 256; i++)
+    for (i = 0; i < 64; i++)
         decoding_table[(unsigned char) encoding_table[i]] = (char)i;
 }
 
@@ -705,74 +787,116 @@ unsigned char *base64_decode(const char *data,
     if (data[input_length - 1] == '=') (*output_length)--;
     if (data[input_length - 2] == '=') (*output_length)--;
 
-    unsigned char *decoded_data = (unsigned char*)malloc(*output_length);
+    unsigned char *decoded_data = (unsigned char*)turn_malloc(*output_length);
     if (decoded_data == NULL) return NULL;
 
-    size_t i,j;
-    for (i = 0, j = 0; i < input_length;) {
+    int i;
+    size_t j;
+    for (i = 0, j = 0; i < (int)input_length;) {
 
-        u32bits sextet_a = 0;
-        if(data[i] != '=')
-        	sextet_a = (u32bits)decoding_table[(int)data[i]];
-        ++i;
-        u32bits sextet_b = 0;
-        if(data[i] != '=')
-        	sextet_b = (u32bits)decoding_table[(int)data[i]];
-        ++i;
-        u32bits sextet_c = 0;
-        if(data[i] != '=')
-        	sextet_c = (u32bits)decoding_table[(int)data[i]];
-        ++i;
-        u32bits sextet_d = 0;
-        if(data[i] != '=')
-        	sextet_c = (u32bits)decoding_table[(int)data[i]];
+		uint32_t sextet_a =
+				data[i] == '=' ? 0 & i++ : decoding_table[(int)data[i++]];
+		uint32_t sextet_b =
+				data[i] == '=' ? 0 & i++ : decoding_table[(int)data[i++]];
+		uint32_t sextet_c =
+				data[i] == '=' ? 0 & i++ : decoding_table[(int)data[i++]];
+		uint32_t sextet_d =
+				data[i] == '=' ? 0 & i++ : decoding_table[(int)data[i++]];
 
-        ++i;
+		uint32_t triple = (sextet_a << 3 * 6) + (sextet_b << 2 * 6)
+				+ (sextet_c << 1 * 6) + (sextet_d << 0 * 6);
 
-        u32bits triple = (sextet_a << 3 * 6)
-        + (sextet_b << 2 * 6)
-        + (sextet_c << 1 * 6)
-        + (sextet_d << 0 * 6);
+		if (j < *output_length)
+			decoded_data[j++] = (triple >> 2 * 8) & 0xFF;
+		if (j < *output_length)
+			decoded_data[j++] = (triple >> 1 * 8) & 0xFF;
+		if (j < *output_length)
+			decoded_data[j++] = (triple >> 0 * 8) & 0xFF;
 
-        if (j < *output_length) decoded_data[j++] = (triple >> 2 * 8) & 0xFF;
-        if (j < *output_length) decoded_data[j++] = (triple >> 1 * 8) & 0xFF;
-        if (j < *output_length) decoded_data[j++] = (triple >> 0 * 8) & 0xFF;
     }
 
     return decoded_data;
 }
 
-/////////////////////////// HMAC ////////////////////////////
+////////////////// SSL /////////////////////
 
-#if defined(__USE_OPENSSL__)
-
-#include <openssl/md5.h>
-#include <openssl/hmac.h>
-#include <openssl/ssl.h>
-
-int calculate_hmac(u08bits *buf, size_t len, const void *key, int key_len, u08bits *hmac, unsigned int *hmac_len)
+static const char* turn_get_method(const SSL_METHOD *method, const char* mdefault)
 {
-	if (!HMAC(EVP_sha1(), key, key_len, buf, len, hmac, hmac_len)) {
-		return -1;
-	} else {
-		return 0;
+	{
+		if(!method)
+			return mdefault;
+		else {
+
+#ifndef OPENSSL_NO_SSL2
+			if(method == SSLv2_server_method()) {
+					return "SSLv2";
+			} else if(method == SSLv2_client_method()) {
+					return "SSLv2";
+			} else
+#endif
+
+			if(method == SSLv3_server_method()) {
+				return "SSLv3";
+			} else if(method == SSLv3_client_method()) {
+				return "SSLv3";
+			} else if(method == SSLv23_server_method()) {
+					return "SSLv23";
+			} else if(method == SSLv23_client_method()) {
+					return "SSLv23";
+			} else if(method == TLSv1_server_method()) {
+					return "TLSv1.0";
+			} else if(method == TLSv1_client_method()) {
+				return "TLSv1.0";
+#if defined(SSL_TXT_TLSV1_1)
+			} else if(method == TLSv1_1_server_method()) {
+					return "TLSv1.1";
+			} else if(method == TLSv1_1_client_method()) {
+				return "TLSv1.1";
+#if defined(SSL_TXT_TLSV1_2)
+			} else if(method == TLSv1_2_server_method()) {
+					return "TLSv1.2";
+			} else if(method == TLSv1_2_client_method()) {
+				return "TLSv1.2";
+#endif
+#endif
+#if !defined(TURN_NO_DTLS)
+			} else if(method == DTLSv1_server_method()) {
+					return "DTLSv1.0";
+			} else if(method == DTLSv1_client_method()) {
+				return "DTLSv1.0";
+#endif
+			} else {
+				if(mdefault)
+					return mdefault;
+				return "UNKNOWN";
+			}
+		}
+	}
+
+}
+
+const char* turn_get_ssl_method(SSL *ssl, const char* mdefault)
+{
+	if(!ssl)
+		return mdefault;
+	else {
+		const SSL_METHOD *method = SSL_get_ssl_method(ssl);
+		if(!method)
+			return mdefault;
+		else
+			return turn_get_method(method, mdefault);
 	}
 }
 
-#else
+//////////// EVENT BASE ///////////////
 
-int calculate_hmac(u08bits *buf, size_t len, const void *key, int key_len, u08bits *hmac, unsigned int *hmac_len)
+struct event_base *turn_event_base_new(void)
 {
-	UNUSED_ARG(buf);
-	UNUSED_ARG(len);
-	UNUSED_ARG(key);
-	UNUSED_ARG(key_len);
-	UNUSED_ARG(hmac);
-	UNUSED_ARG(hmac_len);
+	struct event_config *cfg = event_config_new();
 
-	return -1;
+	event_config_set_flag(cfg,EVENT_BASE_FLAG_EPOLL_USE_CHANGELIST);
+
+	return event_base_new_with_config(cfg);
 }
 
-#endif
-
 //////////////////////////////////////////////////////////////
