@@ -42,6 +42,8 @@
 /////////////////////////////////////////
 
 #define MAX_CONNECT_EFFORTS (77)
+#define DTLS_MAX_CONNECT_TIMEOUT (30)
+#define EXTRA_CREATE_PERMS (25)
 
 static uint64_t current_reservation_token = 0;
 static int allocate_rtcp = 0;
@@ -49,10 +51,24 @@ static const int never_allocate_rtcp = 0;
 
 /////////////////////////////////////////
 
+int rare_event(void)
+{
+	if(dos)
+		return (((unsigned long)random()) %1000 == 777);
+	return 0;
+}
+
+int not_rare_event(void)
+{
+	if(dos)
+		return ((((unsigned long)random()) %1000) < 200);
+	return 0;
+}
+
 static int get_allocate_address_family(ioa_addr *relay_addr) {
-	if(relay_addr->ss.ss_family == AF_INET)
+	if(relay_addr->ss.sa_family == AF_INET)
 		return STUN_ATTRIBUTE_REQUESTED_ADDRESS_FAMILY_VALUE_DEFAULT;
-	else if(relay_addr->ss.ss_family == AF_INET6)
+	else if(relay_addr->ss.sa_family == AF_INET6)
 		return STUN_ATTRIBUTE_REQUESTED_ADDRESS_FAMILY_VALUE_IPV6;
 	else
 		return STUN_ATTRIBUTE_REQUESTED_ADDRESS_FAMILY_VALUE_INVALID;
@@ -62,7 +78,9 @@ static int get_allocate_address_family(ioa_addr *relay_addr) {
 
 static SSL* tls_connect(ioa_socket_raw fd, ioa_addr *remote_addr)
 {
-	SSL *ssl = SSL_new(root_tls_ctx);
+	int ctxtype = (int)(((unsigned long)random())%root_tls_ctx_num);
+
+	SSL *ssl = SSL_new(root_tls_ctx[ctxtype]);
 
 	if(use_tcp) {
 		SSL_set_fd(ssl, fd);
@@ -88,7 +106,7 @@ static SSL* tls_connect(ioa_socket_raw fd, ioa_addr *remote_addr)
 			BIO_ctrl(bio, BIO_CTRL_DGRAM_SET_RECV_TIMEOUT, 0, &timeout);
 		}
 
-		set_mtu_df(ssl, fd, remote_addr->ss.ss_family, SOSO_MTU, !use_tcp, clnet_verbose);
+		set_mtu_df(ssl, fd, remote_addr->ss.sa_family, SOSO_MTU, !use_tcp, clnet_verbose);
 #endif
 	}
 
@@ -104,14 +122,23 @@ static SSL* tls_connect(ioa_socket_raw fd, ioa_addr *remote_addr)
 			rc = SSL_connect(ssl);
 		} while (rc < 0 && errno == EINTR);
 		if (rc > 0) {
-			break;
+		  TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,"%s: client session connected with cipher %s, method=%s\n",__FUNCTION__,
+				  SSL_get_cipher(ssl),turn_get_ssl_method(ssl,NULL));
+		  if(clnet_verbose && SSL_get_peer_certificate(ssl)) {
+			  TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "------------------------------------------------------------\n");
+		  	X509_NAME_print_ex_fp(stdout, X509_get_subject_name(SSL_get_peer_certificate(ssl)), 1,
+		  						XN_FLAG_MULTILINE);
+		  	TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "\n\n Cipher: %s\n", SSL_CIPHER_get_name(SSL_get_current_cipher(ssl)));
+		  	TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "\n------------------------------------------------------------\n\n");
+		  }
+		  break;
 		} else {
 			TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "%s: cannot connect\n",
 					__FUNCTION__);
 			switch (SSL_get_error(ssl, rc)) {
 			case SSL_ERROR_WANT_READ:
 			case SSL_ERROR_WANT_WRITE:
-				usleep(1000);
+				if(!dos) usleep(1000);
 				continue;
 			default: {
 				char buf[1025];
@@ -178,16 +205,14 @@ static int clnet_connect(uint16_t clnet_remote_port, const char *remote_address,
 			&remote_addr) < 0)
 		return -1;
 
-	ns_bzero(&local_addr, sizeof(struct sockaddr_storage));
+	ns_bzero(&local_addr, sizeof(ioa_addr));
 
-	clnet_fd = socket(remote_addr.ss.ss_family, use_tcp ? SOCK_STREAM : SOCK_DGRAM, 0);
+	clnet_fd = socket(remote_addr.ss.sa_family, use_tcp ? SOCK_STREAM : SOCK_DGRAM, 0);
 	if (clnet_fd < 0) {
 		perror("socket");
 		exit(-1);
 	}
 
-	socket_set_reusable(clnet_fd);
-
 	if (sock_bind_to_device(clnet_fd, ifname) < 0) {
 		TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,
 				"Cannot bind client socket to device %s\n", ifname);
@@ -196,18 +221,26 @@ static int clnet_connect(uint16_t clnet_remote_port, const char *remote_address,
 	set_sock_buf_size(clnet_fd, UR_CLIENT_SOCK_BUF_SIZE);
 
 	if(clnet_info->is_peer && (*local_address==0)) {
-		if (make_ioa_addr((const u08bits*) "127.0.0.1", 0,
-					    &local_addr) < 0)
-			    return -1;
 
-		addr_bind(clnet_fd, &local_addr);
+		if(remote_addr.ss.sa_family == AF_INET6) {
+			if (make_ioa_addr((const u08bits*) "::1", 0, &local_addr) < 0) {
+			    return -1;
+			}
+		} else {
+			if (make_ioa_addr((const u08bits*) "127.0.0.1", 0, &local_addr) < 0) {
+			    return -1;
+			}
+		}
+
+		addr_bind(clnet_fd, &local_addr, 0);
+
 	} else if (strlen(local_address) > 0) {
 
-	  if (make_ioa_addr((const u08bits*) local_address, 0,
+		if (make_ioa_addr((const u08bits*) local_address, 0,
 			    &local_addr) < 0)
-	    return -1;
+			return -1;
 
-	  addr_bind(clnet_fd, &local_addr);
+		addr_bind(clnet_fd, &local_addr,0);
 	}
 
 	if(clnet_info->is_peer) {
@@ -220,9 +253,9 @@ static int clnet_connect(uint16_t clnet_remote_port, const char *remote_address,
 		addr_cpy(&(clnet_info->local_addr), &local_addr);
 		clnet_info->fd = clnet_fd;
 		addr_get_from_sock(clnet_fd, &(clnet_info->local_addr));
-		strcpy(clnet_info->lsaddr,local_address);
-		strcpy(clnet_info->rsaddr,remote_address);
-		strcpy(clnet_info->ifname,(const char*)ifname);
+		STRCPY(clnet_info->lsaddr,local_address);
+		STRCPY(clnet_info->rsaddr,remote_address);
+		STRCPY(clnet_info->ifname,(const char*)ifname);
 	}
 
 	if (use_secure) {
@@ -233,13 +266,42 @@ static int clnet_connect(uint16_t clnet_remote_port, const char *remote_address,
 		}
 	}
 
-	addr_debug_print(verbose, &remote_addr, "Connected to");
+	if(verbose && clnet_info) {
+		addr_debug_print(verbose, &(clnet_info->local_addr), "Connected from");
+		addr_debug_print(verbose, &remote_addr, "Connected to");
+	}
 
-	usleep(50000);
+	if(!dos) usleep(500);
 
 	return 0;
 }
 
+int read_mobility_ticket(app_ur_conn_info *clnet_info, stun_buffer *message)
+{
+	int ret = 0;
+	if(clnet_info && message) {
+		stun_attr_ref s_mobile_id_sar = stun_attr_get_first_by_type(message, STUN_ATTRIBUTE_MOBILITY_TICKET);
+		if(s_mobile_id_sar) {
+			int smid_len = stun_attr_get_len(s_mobile_id_sar);
+			if(smid_len>0 && (((size_t)smid_len)<sizeof(clnet_info->s_mobile_id))) {
+				const u08bits* smid_val = stun_attr_get_value(s_mobile_id_sar);
+				if(smid_val) {
+					ns_bcopy(smid_val, clnet_info->s_mobile_id, (size_t)smid_len);
+					clnet_info->s_mobile_id[smid_len] = 0;
+					if (clnet_verbose)
+						TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,
+									"%s: smid=%s\n", __FUNCTION__, clnet_info->s_mobile_id);
+				}
+			} else {
+				TURN_LOG_FUNC(TURN_LOG_LEVEL_ERROR,
+							"%s: ERROR: smid_len=%d\n", __FUNCTION__, smid_len);
+				ret = -1;
+			}
+		}
+	}
+	return ret;
+}
+
 static int clnet_allocate(int verbose,
 		app_ur_conn_info *clnet_info,
 		ioa_addr *relay_addr,
@@ -260,7 +322,7 @@ static int clnet_allocate(int verbose,
 		int allocate_sent = 0;
 
 		if(reopen_socket && !use_tcp) {
-			close(clnet_info->fd);
+			socket_closesocket(clnet_info->fd);
 			clnet_info->fd = -1;
 			if (clnet_connect(addr_get_port(&(clnet_info->remote_addr)), clnet_info->rsaddr, (u08bits*)clnet_info->ifname, clnet_info->lsaddr,
 					verbose, clnet_info) < 0) {
@@ -272,7 +334,11 @@ static int clnet_allocate(int verbose,
 		stun_buffer message;
 		if(current_reservation_token)
 			af = STUN_ATTRIBUTE_REQUESTED_ADDRESS_FAMILY_VALUE_DEFAULT;
-		stun_set_allocate_request(&message, 1800, af, relay_transport);
+
+		if(!dos)
+			stun_set_allocate_request(&message, 800, af, relay_transport, mobility);
+		else
+			stun_set_allocate_request(&message, 300, af, relay_transport, mobility);
 		if(dont_fragment)
 			stun_attr_add(&message, STUN_ATTRIBUTE_DONT_FRAGMENT, NULL, 0);
 		if(!no_rtcp) {
@@ -306,6 +372,8 @@ static int clnet_allocate(int verbose,
 
 		////////////<<==allocate send
 
+		if(not_rare_event()) return 0;
+
 		////////allocate response==>>
 		{
 			int allocate_received = 0;
@@ -327,9 +395,10 @@ static int clnet_allocate(int verbose,
 						allocate_finished = 1;
 
 						if(clnet_info->nonce[0] || use_short_term) {
+							SHATYPE sht = clnet_info->shatype;
 							if(stun_check_message_integrity_str(get_turn_credentials_type(),
 											message.buf, (size_t)(message.len), g_uname,
-										clnet_info->realm, g_upwd)<1) {
+										clnet_info->realm, g_upwd, sht)<1) {
 								TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,"Wrong integrity in allocate message received from server\n");
 								return -1;
 							}
@@ -349,7 +418,7 @@ static int clnet_allocate(int verbose,
 							if (verbose) {
 								ioa_addr remote_addr;
 								memcpy(&remote_addr, relay_addr,
-										sizeof(remote_addr));
+										sizeof(ioa_addr));
 								addr_debug_print(verbose, &remote_addr,
 										"Received relay addr");
 							}
@@ -361,20 +430,33 @@ static int clnet_allocate(int verbose,
 						if (verbose)
 							TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,
 									"%s: rtv=%llu\n", __FUNCTION__, rtv);
+
+						read_mobility_ticket(clnet_info, &message);
+
 					} else if (stun_is_challenge_response_str(message.buf, (size_t)message.len,
 									&err_code,err_msg,sizeof(err_msg),
 									clnet_info->realm,clnet_info->nonce)) {
+						if(err_code == SHA_TOO_WEAK && (clnet_info->shatype == SHATYPE_SHA1)) {
+							clnet_info->shatype = SHATYPE_SHA256;
+							recalculate_restapi_hmac();
+						}
 						goto beg_allocate;
 					} else if (stun_is_error_response(&message, &err_code,err_msg,sizeof(err_msg))) {
 
+						if(err_code == SHA_TOO_WEAK && (clnet_info->shatype == SHATYPE_SHA1) && use_short_term) {
+							clnet_info->shatype = SHATYPE_SHA256;
+							goto beg_allocate;
+						}
+
 						allocate_received = 1;
 
 						if(err_code == 300) {
 
 							if(clnet_info->nonce[0] || use_short_term) {
+								SHATYPE sht = clnet_info->shatype;
 								if(stun_check_message_integrity_str(get_turn_credentials_type(),
 											message.buf, (size_t)(message.len), g_uname,
-											clnet_info->realm, g_upwd)<1) {
+											clnet_info->realm, g_upwd, sht)<1) {
 									TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,"Wrong integrity in allocate message received from server\n");
 									return -1;
 								}
@@ -417,6 +499,8 @@ static int clnet_allocate(int verbose,
 	}
 	////////////<<== allocate response received
 
+	if(rare_event()) return 0;
+
 	if(!allocate_finished) {
 	  TURN_LOG_FUNC(TURN_LOG_LEVEL_ERROR,
 			"Cannot complete Allocation\n");
@@ -429,6 +513,48 @@ static int clnet_allocate(int verbose,
 
 	  af_cycle = 0;
 
+	  if(clnet_info->s_mobile_id[0]) {
+
+		  int fd = clnet_info->fd;
+		  SSL* ssl = clnet_info->ssl;
+
+		  int close_now = (int)(random()%2);
+
+		  if(close_now) {
+			  int close_socket = (int)(random()%2);
+			  if(ssl && !close_socket) {
+				  SSL_shutdown(ssl);
+				  SSL_free(ssl);
+				  ssl = NULL;
+				  fd = -1;
+			  } else if(fd>=0) {
+				  close(fd);
+				  fd = -1;
+				  ssl = NULL;
+			  }
+		  }
+
+		  app_ur_conn_info ci;
+		  ns_bcopy(clnet_info,&ci,sizeof(ci));
+		  ci.fd = -1;
+		  ci.ssl = NULL;
+		  clnet_info->fd = -1;
+		  clnet_info->ssl = NULL;
+		  //Reopen:
+		  if(clnet_connect(addr_get_port(&(ci.remote_addr)), ci.rsaddr,
+		  		(unsigned char*)ci.ifname, ci.lsaddr, clnet_verbose,
+		  		clnet_info)<0) {
+			  exit(-1);
+		  }
+
+		  if(ssl) {
+			  SSL_shutdown(ssl);
+		  	  SSL_free(ssl);
+		  } else if(fd>=0) {
+		  	  close(fd);
+		  }
+	  }
+
 		beg_refresh:
 
 	  if(af_cycle++>32) {
@@ -447,6 +573,10 @@ static int clnet_allocate(int verbose,
 			stun_attr_add(&message, STUN_ATTRIBUTE_LIFETIME, (const char*) &lt,
 					4);
 
+			if(clnet_info->s_mobile_id[0]) {
+				stun_attr_add(&message, STUN_ATTRIBUTE_MOBILITY_TICKET, (const char*)clnet_info->s_mobile_id, strlen(clnet_info->s_mobile_id));
+			}
+
 			if(add_integrity(clnet_info, &message)<0) return -1;
 
 			stun_attr_add_fingerprint_str(message.buf,(size_t*)&(message.len));
@@ -467,6 +597,8 @@ static int clnet_allocate(int verbose,
 			}
 		}
 
+		if(not_rare_event()) return 0;
+
 		////////refresh response==>>
 		{
 			int refresh_received = 0;
@@ -484,6 +616,7 @@ static int clnet_allocate(int verbose,
 					int err_code = 0;
 					u08bits err_msg[129];
 					if (stun_is_success_response(&message)) {
+						read_mobility_ticket(clnet_info, &message);
 						refresh_received = 1;
 						if (verbose) {
 							TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "success\n");
@@ -491,7 +624,11 @@ static int clnet_allocate(int verbose,
 					} else if (stun_is_challenge_response_str(message.buf, (size_t)message.len,
 										&err_code,err_msg,sizeof(err_msg),
 										clnet_info->realm,clnet_info->nonce)) {
-							goto beg_refresh;
+						if(err_code == SHA_TOO_WEAK && (clnet_info->shatype == SHATYPE_SHA1)) {
+							clnet_info->shatype = SHATYPE_SHA256;
+							recalculate_restapi_hmac();
+						}
+						goto beg_refresh;
 					} else if (stun_is_error_response(&message, &err_code,err_msg,sizeof(err_msg))) {
 						refresh_received = 1;
 						TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "error %d (%s)\n",
@@ -523,7 +660,11 @@ static int turn_channel_bind(int verbose, uint16_t *chn,
 
 		stun_buffer message;
 
-		*chn = stun_set_channel_bind_request(&message, peer_addr, *chn);
+		if(negative_test) {
+			*chn = stun_set_channel_bind_request(&message, peer_addr, (u16bits)random());
+		} else {
+			*chn = stun_set_channel_bind_request(&message, peer_addr, *chn);
+		}
 
 		if(add_integrity(clnet_info, &message)<0) return -1;
 
@@ -546,6 +687,8 @@ static int turn_channel_bind(int verbose, uint16_t *chn,
 
 	////////////<<==channel bind send
 
+	if(not_rare_event()) return 0;
+
 	////////channel bind response==>>
 
 	{
@@ -566,9 +709,10 @@ static int turn_channel_bind(int verbose, uint16_t *chn,
 					cb_received = 1;
 
 					if(clnet_info->nonce[0] || use_short_term) {
+						SHATYPE sht = clnet_info->shatype;
 						if(stun_check_message_integrity_str(get_turn_credentials_type(),
 										message.buf, (size_t)(message.len), g_uname,
-									clnet_info->realm, g_upwd)<1) {
+									clnet_info->realm, g_upwd, sht)<1) {
 							TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,"Wrong integrity in channel bind message received from server\n");
 							return -1;
 						}
@@ -581,7 +725,11 @@ static int turn_channel_bind(int verbose, uint16_t *chn,
 				} else if (stun_is_challenge_response_str(message.buf, (size_t)message.len,
 										&err_code,err_msg,sizeof(err_msg),
 										clnet_info->realm,clnet_info->nonce)) {
-										goto beg_bind;
+					if(err_code == SHA_TOO_WEAK && (clnet_info->shatype == SHATYPE_SHA1)) {
+						clnet_info->shatype = SHATYPE_SHA256;
+						recalculate_restapi_hmac();
+					}
+					goto beg_bind;
 				} else if (stun_is_error_response(&message, &err_code,err_msg,sizeof(err_msg))) {
 					cb_received = 1;
 					TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "channel bind: error %d (%s)\n",
@@ -603,7 +751,16 @@ static int turn_channel_bind(int verbose, uint16_t *chn,
 }
 
 static int turn_create_permission(int verbose, app_ur_conn_info *clnet_info,
-		ioa_addr *peer_addr) {
+		ioa_addr *peer_addr, int addrnum)
+{
+
+	if(no_permissions || (addrnum<1))
+		return 0;
+
+	char saddr[129]="\0";
+	if (verbose) {
+		addr_to_string(peer_addr,(u08bits*)saddr);
+	}
 
 	beg_cp:
 
@@ -613,7 +770,12 @@ static int turn_create_permission(int verbose, app_ur_conn_info *clnet_info,
 		stun_buffer message;
 
 		stun_init_request(STUN_METHOD_CREATE_PERMISSION, &message);
-		stun_attr_add_addr(&message, STUN_ATTRIBUTE_XOR_PEER_ADDRESS, peer_addr);
+		{
+			int addrindex;
+			for(addrindex=0;addrindex<addrnum;++addrindex) {
+				stun_attr_add_addr(&message, STUN_ATTRIBUTE_XOR_PEER_ADDRESS, peer_addr+addrindex);
+			}
+		}
 
 		if(add_integrity(clnet_info, &message)<0) return -1;
 
@@ -625,7 +787,7 @@ static int turn_create_permission(int verbose, app_ur_conn_info *clnet_info,
 
 			if (len > 0) {
 				if (verbose) {
-					TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "create perm sent\n");
+					TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "create perm sent: %s\n",saddr);
 				}
 				cp_sent = 1;
 			} else {
@@ -637,6 +799,8 @@ static int turn_create_permission(int verbose, app_ur_conn_info *clnet_info,
 
 	////////////<<==create permission send
 
+	if(not_rare_event()) return 0;
+
 	////////create permission response==>>
 
 	{
@@ -657,9 +821,10 @@ static int turn_create_permission(int verbose, app_ur_conn_info *clnet_info,
 					cp_received = 1;
 
 					if(clnet_info->nonce[0] || use_short_term) {
+						SHATYPE sht = clnet_info->shatype;
 						if(stun_check_message_integrity_str(get_turn_credentials_type(),
 										message.buf, (size_t)(message.len), g_uname,
-											clnet_info->realm, g_upwd)<1) {
+											clnet_info->realm, g_upwd, sht)<1) {
 							TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,"Wrong integrity in create permission message received from server\n");
 							return -1;
 						}
@@ -671,6 +836,10 @@ static int turn_create_permission(int verbose, app_ur_conn_info *clnet_info,
 				} else if (stun_is_challenge_response_str(message.buf, (size_t)message.len,
 									&err_code,err_msg,sizeof(err_msg),
 									clnet_info->realm,clnet_info->nonce)) {
+					if(err_code == SHA_TOO_WEAK && (clnet_info->shatype == SHATYPE_SHA1)) {
+						clnet_info->shatype = SHATYPE_SHA256;
+						recalculate_restapi_hmac();
+					}
 					goto beg_cp;
 				} else if (stun_is_error_response(&message, &err_code,err_msg,sizeof(err_msg))) {
 					cp_received = 1;
@@ -738,54 +907,158 @@ int start_connection(uint16_t clnet_remote_port0,
 	  }
 	}
 
-	if (clnet_allocate(verbose, clnet_info, &relay_addr, get_allocate_address_family(&peer_addr),NULL,NULL) < 0) {
+	int af = default_address_family ? default_address_family : get_allocate_address_family(&peer_addr);
+	if (clnet_allocate(verbose, clnet_info, &relay_addr, af, NULL,NULL) < 0) {
 	  exit(-1);
 	}
 
+	if(rare_event()) return 0;
+
 	if(!no_rtcp) {
-	  if (clnet_allocate(verbose, clnet_info_rtcp, &relay_addr_rtcp, get_allocate_address_family(&peer_addr_rtcp),NULL,NULL) < 0) {
+		af = default_address_family ? default_address_family : get_allocate_address_family(&peer_addr_rtcp);
+	  if (clnet_allocate(verbose, clnet_info_rtcp, &relay_addr_rtcp, af,NULL,NULL) < 0) {
 	    exit(-1);
 	  }
+	  if(rare_event()) return 0;
 	}
 
-	if (!do_not_use_channel) {
+	if (!dos) {
+		if (!do_not_use_channel) {
+			/* These multiple "channel bind" requests are here only because
+			 * we are playing with the TURN server trying to screw it */
+			if (turn_channel_bind(verbose, chn, clnet_info, &peer_addr_rtcp)
+					< 0) {
+				exit(-1);
+			}
+			if(rare_event()) return 0;
 
-		/* These multiple "channel bind" requests are here only because
-		 * we are playing with the TURN server trying to screw it */
-		if (turn_channel_bind(verbose, chn, clnet_info, &peer_addr_rtcp) < 0) {
-			exit(-1);
-		}
-		if (turn_channel_bind(verbose, chn, clnet_info, &peer_addr_rtcp) < 0) {
-			exit(-1);
-		}
-		*chn = 0;
-		if (turn_channel_bind(verbose, chn, clnet_info, &peer_addr) < 0) {
-			exit(-1);
-		}
-		if (turn_channel_bind(verbose, chn, clnet_info, &peer_addr) < 0) {
-			exit(-1);
-		}
-		if(!no_rtcp) {
-		  if (turn_channel_bind(verbose, chn_rtcp, clnet_info_rtcp, &peer_addr_rtcp) < 0) {
-		    exit(-1);
-		  }
-		}
-	} else {
-		if (turn_create_permission(verbose, clnet_info, &peer_addr) < 0) {
-			exit(-1);
-		}
-		if (turn_create_permission(verbose, clnet_info, &peer_addr_rtcp) < 0) {
-			exit(-1);
-		}
-		if(!no_rtcp) {
-		  if (turn_create_permission(verbose, clnet_info_rtcp, &peer_addr_rtcp)
-		      < 0) {
-		    exit(-1);
-		  }
-		  if (turn_create_permission(verbose, clnet_info_rtcp, &peer_addr)
-		      < 0) {
-		    exit(-1);
-		  }
+			if (turn_channel_bind(verbose, chn, clnet_info, &peer_addr_rtcp)
+					< 0) {
+				exit(-1);
+			}
+			if(rare_event()) return 0;
+			*chn = 0;
+			if (turn_channel_bind(verbose, chn, clnet_info, &peer_addr) < 0) {
+				exit(-1);
+			}
+
+			if(rare_event()) return 0;
+			if (turn_channel_bind(verbose, chn, clnet_info, &peer_addr) < 0) {
+				exit(-1);
+			}
+			if(rare_event()) return 0;
+
+			if(extra_requests) {
+				const char *sarbaddr = "164.156.178.190";
+				if(random() % 2 == 0)
+					sarbaddr = "2001::172";
+				ioa_addr arbaddr;
+				make_ioa_addr((const u08bits*)sarbaddr, 333, &arbaddr);
+				int i;
+				int maxi = (unsigned short)random() % EXTRA_CREATE_PERMS;
+				for(i=0;i<maxi;i++) {
+					u16bits chni=0;
+					int port = (unsigned short)random();
+					if(port<1024) port += 1024;
+					addr_set_port(&arbaddr, port);
+					u08bits *u=(u08bits*)&(arbaddr.s4.sin_addr);
+					u[(unsigned short)random()%4] = u[(unsigned short)random()%4] + 1;
+					//char sss[128];
+					//addr_to_string(&arbaddr,(u08bits*)sss);
+					//printf("%s: 111.111: %s\n",__FUNCTION__,sss);
+					turn_channel_bind(verbose, &chni, clnet_info, &arbaddr);
+				}
+			}
+
+			if (!no_rtcp) {
+				if (turn_channel_bind(verbose, chn_rtcp, clnet_info_rtcp,
+						&peer_addr_rtcp) < 0) {
+					exit(-1);
+				}
+			}
+			if(rare_event()) return 0;
+
+			if(extra_requests) {
+				const char *sarbaddr = "64.56.78.90";
+				if(random() % 2 == 0)
+					sarbaddr = "2001::172";
+				ioa_addr arbaddr[EXTRA_CREATE_PERMS];
+				make_ioa_addr((const u08bits*)sarbaddr, 333, &arbaddr[0]);
+				int i;
+				int maxi = (unsigned short)random() % EXTRA_CREATE_PERMS;
+				for(i=0;i<maxi;i++) {
+					if(i>0)
+						addr_cpy(&arbaddr[i],&arbaddr[0]);
+					addr_set_port(&arbaddr[i], (unsigned short)random());
+					u08bits *u=(u08bits*)&(arbaddr[i].s4.sin_addr);
+					u[(unsigned short)random()%4] = u[(unsigned short)random()%4] + 1;
+					//char sss[128];
+					//addr_to_string(&arbaddr[i],(u08bits*)sss);
+					//printf("%s: 111.111: %s\n",__FUNCTION__,sss);
+				}
+				turn_create_permission(verbose, clnet_info, arbaddr, maxi);
+			}
+		} else {
+
+			int before=(random()%2 == 0);
+
+			if(before) {
+				if (turn_create_permission(verbose, clnet_info, &peer_addr, 1) < 0) {
+					exit(-1);
+				}
+				if(rare_event()) return 0;
+				if (turn_create_permission(verbose, clnet_info, &peer_addr_rtcp, 1)
+						< 0) {
+					exit(-1);
+				}
+				if(rare_event()) return 0;
+			}
+
+			if(extra_requests) {
+				const char *sarbaddr = "64.56.78.90";
+				if(random() % 2 == 0)
+					sarbaddr = "2001::172";
+				ioa_addr arbaddr[EXTRA_CREATE_PERMS];
+				make_ioa_addr((const u08bits*)sarbaddr, 333, &arbaddr[0]);
+				int i;
+				int maxi = (unsigned short)random() % EXTRA_CREATE_PERMS;
+				for(i=0;i<maxi;i++) {
+					if(i>0)
+						addr_cpy(&arbaddr[i],&arbaddr[0]);
+					addr_set_port(&arbaddr[i], (unsigned short)random());
+					u08bits *u=(u08bits*)&(arbaddr[i].s4.sin_addr);
+					u[(unsigned short)random()%4] = u[(unsigned short)random()%4] + 1;
+					//char sss[128];
+					//addr_to_string(&arbaddr,(u08bits*)sss);
+					//printf("%s: 111.111: %s\n",__FUNCTION__,sss);
+				}
+				turn_create_permission(verbose, clnet_info, arbaddr, maxi);
+			}
+
+			if(!before) {
+				if (turn_create_permission(verbose, clnet_info, &peer_addr, 1) < 0) {
+					exit(-1);
+				}
+				if(rare_event()) return 0;
+				if (turn_create_permission(verbose, clnet_info, &peer_addr_rtcp, 1)
+					< 0) {
+					exit(-1);
+				}
+				if(rare_event()) return 0;
+			}
+
+			if (!no_rtcp) {
+				if (turn_create_permission(verbose, clnet_info_rtcp,
+						&peer_addr_rtcp, 1) < 0) {
+					exit(-1);
+				}
+				if(rare_event()) return 0;
+				if (turn_create_permission(verbose, clnet_info_rtcp, &peer_addr, 1)
+						< 0) {
+					exit(-1);
+				}
+				if(rare_event()) return 0;
+			}
 		}
 	}
 
@@ -832,6 +1105,8 @@ int start_c2c_connection(uint16_t clnet_remote_port0,
 
 	clnet_allocate(verbose, clnet_info_probe, &relay_addr1, default_address_family, remote_address, &clnet_remote_port);
 
+	if(rare_event()) return 0;
+
 	/* Real: */
 
 	if (clnet_connect(clnet_remote_port, remote_address, ifname, local_address,
@@ -866,30 +1141,40 @@ int start_c2c_connection(uint16_t clnet_remote_port0,
 	    exit(-1);
 	  }
 	  
+	  if(rare_event()) return 0;
+
 	  if (clnet_allocate(verbose, clnet_info1_rtcp,
 			   &relay_addr1_rtcp, default_address_family,NULL,NULL) < 0) {
 	    exit(-1);
 	  }
 	  
+	  if(rare_event()) return 0;
+
 	  if (clnet_allocate(verbose, clnet_info2, &relay_addr2, default_address_family,NULL,NULL)
 	      < 0) {
 	    exit(-1);
 	  }
 	  
+	  if(rare_event()) return 0;
+
 	  if (clnet_allocate(verbose, clnet_info2_rtcp,
 			   &relay_addr2_rtcp, default_address_family,NULL,NULL) < 0) {
 	    exit(-1);
 	  }
+
+	  if(rare_event()) return 0;
 	} else {
 
 	  if (clnet_allocate(verbose, clnet_info1, &relay_addr1, default_address_family,NULL,NULL)
 	      < 0) {
 	    exit(-1);
 	  }
+	  if(rare_event()) return 0;
 	  if(!(clnet_info2->is_peer)) {
 		  if (clnet_allocate(verbose, clnet_info2, &relay_addr2, default_address_family,NULL,NULL) < 0) {
 			  exit(-1);
 		  }
+		  if(rare_event()) return 0;
 	  } else {
 		  addr_cpy(&(clnet_info2->remote_addr),&relay_addr1);
 		  addr_cpy(&relay_addr2,&(clnet_info2->local_addr));
@@ -900,37 +1185,110 @@ int start_c2c_connection(uint16_t clnet_remote_port0,
 		if (turn_channel_bind(verbose, chn1, clnet_info1, &relay_addr2) < 0) {
 			exit(-1);
 		}
+
+		if(extra_requests) {
+			const char *sarbaddr = "164.156.178.190";
+			if(random() % 2 == 0)
+				sarbaddr = "2001::172";
+			ioa_addr arbaddr;
+			make_ioa_addr((const u08bits*)sarbaddr, 333, &arbaddr);
+			int i;
+			int maxi = (unsigned short)random() % EXTRA_CREATE_PERMS;
+			for(i=0;i<maxi;i++) {
+				u16bits chni=0;
+				int port = (unsigned short)random();
+				if(port<1024) port += 1024;
+				addr_set_port(&arbaddr, port);
+				u08bits *u=(u08bits*)&(arbaddr.s4.sin_addr);
+				u[(unsigned short)random()%4] = u[(unsigned short)random()%4] + 1;
+				//char sss[128];
+				//addr_to_string(&arbaddr,(u08bits*)sss);
+				//printf("%s: 111.111: %s\n",__FUNCTION__,sss);
+				turn_channel_bind(verbose, &chni, clnet_info1, &arbaddr);
+			}
+		}
+
+		if(rare_event()) return 0;
+
+		if(extra_requests) {
+			const char *sarbaddr = "64.56.78.90";
+			if(random() % 2 == 0)
+				sarbaddr = "2001::172";
+			ioa_addr arbaddr[EXTRA_CREATE_PERMS];
+			make_ioa_addr((const u08bits*)sarbaddr, 333, &arbaddr[0]);
+			int i;
+			int maxi = (unsigned short)random() % EXTRA_CREATE_PERMS;
+			for(i=0;i<maxi;i++) {
+				if(i>0)
+					addr_cpy(&arbaddr[i],&arbaddr[0]);
+				addr_set_port(&arbaddr[i], (unsigned short)random());
+				u08bits *u=(u08bits*)&(arbaddr[i].s4.sin_addr);
+				u[(unsigned short)random()%4] = u[(unsigned short)random()%4] + 1;
+				//char sss[128];
+				//addr_to_string(&arbaddr[i],(u08bits*)sss);
+				//printf("%s: 111.111: %s\n",__FUNCTION__,sss);
+			}
+			turn_create_permission(verbose, clnet_info1, arbaddr, maxi);
+		}
+
 		if(!no_rtcp)
 		  if (turn_channel_bind(verbose, chn1_rtcp, clnet_info1_rtcp,
 					&relay_addr2_rtcp) < 0) {
 		    exit(-1);
 		  }
+		if(rare_event()) return 0;
 		if (turn_channel_bind(verbose, chn2, clnet_info2, &relay_addr1) < 0) {
 			exit(-1);
 		}
+		if(rare_event()) return 0;
 		if(!no_rtcp)
 		  if (turn_channel_bind(verbose, chn2_rtcp, clnet_info2_rtcp,
 					&relay_addr1_rtcp) < 0) {
 		    exit(-1);
 		  }
+		if(rare_event()) return 0;
 	} else {
 
-		if (turn_create_permission(verbose, clnet_info1, &relay_addr2) < 0) {
+		if (turn_create_permission(verbose, clnet_info1, &relay_addr2, 1) < 0) {
 			exit(-1);
 		}
-		if (!no_rtcp)
-			if (turn_create_permission(verbose, clnet_info1_rtcp, &relay_addr2_rtcp) < 0) {
-				exit(-1);
-			}
-		if(!(clnet_info2->is_peer)) {
-			if (turn_create_permission(verbose, clnet_info2, &relay_addr1) < 0) {
-				exit(-1);
+
+		if(extra_requests) {
+			const char *sarbaddr = "64.56.78.90";
+			if(random() % 2 == 0)
+				sarbaddr = "2001::172";
+			ioa_addr arbaddr;
+			make_ioa_addr((const u08bits*)sarbaddr, 333, &arbaddr);
+			int i;
+			int maxi = (unsigned short)random() % EXTRA_CREATE_PERMS;
+			for(i=0;i<maxi;i++) {
+				addr_set_port(&arbaddr, (unsigned short)random());
+				u08bits *u=(u08bits*)&(arbaddr.s4.sin_addr);
+				u[(unsigned short)random()%4] = u[(unsigned short)random()%4] + 1;
+				//char sss[128];
+				//addr_to_string(&arbaddr,(u08bits*)sss);
+				//printf("%s: 111.111: %s\n",__FUNCTION__,sss);
+				turn_create_permission(verbose, clnet_info1, &arbaddr, 1);
 			}
 		}
+
+		if(rare_event()) return 0;
 		if (!no_rtcp)
-			if (turn_create_permission(verbose, clnet_info2_rtcp, &relay_addr1_rtcp) < 0) {
+			if (turn_create_permission(verbose, clnet_info1_rtcp, &relay_addr2_rtcp, 1) < 0) {
 				exit(-1);
 			}
+		if(rare_event()) return 0;
+		if(!(clnet_info2->is_peer)) {
+			if (turn_create_permission(verbose, clnet_info2, &relay_addr1, 1) < 0) {
+				exit(-1);
+			}
+			if(rare_event()) return 0;
+		}
+		if (!no_rtcp)
+			if (turn_create_permission(verbose, clnet_info2_rtcp, &relay_addr1_rtcp, 1) < 0) {
+				exit(-1);
+			}
+		if(rare_event()) return 0;
 	}
 
 	addr_cpy(&(clnet_info1->peer_addr), &relay_addr2);
@@ -980,7 +1338,7 @@ int turn_tcp_connect(int verbose, app_ur_conn_info *clnet_info, ioa_addr *peer_a
 	return 0;
 }
 
-static int turn_tcp_connection_bind(int verbose, app_ur_conn_info *clnet_info, app_tcp_conn_info *atc) {
+static int turn_tcp_connection_bind(int verbose, app_ur_conn_info *clnet_info, app_tcp_conn_info *atc, int errorOK) {
 
 	beg_cb:
 
@@ -988,9 +1346,11 @@ static int turn_tcp_connection_bind(int verbose, app_ur_conn_info *clnet_info, a
 		int cb_sent = 0;
 
 		stun_buffer message;
+		u32bits cid = atc->cid;
 
 		stun_init_request(STUN_METHOD_CONNECTION_BIND, &message);
-		stun_attr_add(&message, STUN_ATTRIBUTE_CONNECTION_ID, (const s08bits*)&(atc->cid),4);
+
+		stun_attr_add(&message, STUN_ATTRIBUTE_CONNECTION_ID, (const s08bits*)&cid,4);
 
 		if(add_integrity(clnet_info, &message)<0) return -1;
 
@@ -1006,6 +1366,8 @@ static int turn_tcp_connection_bind(int verbose, app_ur_conn_info *clnet_info, a
 				}
 				cb_sent = 1;
 			} else {
+				if(errorOK)
+					return 0;
 				perror("send");
 				exit(1);
 			}
@@ -1014,6 +1376,8 @@ static int turn_tcp_connection_bind(int verbose, app_ur_conn_info *clnet_info, a
 
 	////////////<<==connection bind send
 
+	if(not_rare_event()) return 0;
+
 	////////connection bind response==>>
 
 	{
@@ -1032,9 +1396,10 @@ static int turn_tcp_connection_bind(int verbose, app_ur_conn_info *clnet_info, a
 				if (stun_is_success_response(&message)) {
 
 					if(clnet_info->nonce[0] || use_short_term) {
+						SHATYPE sht = clnet_info->shatype;
 						if(stun_check_message_integrity_str(get_turn_credentials_type(),
 										message.buf, (size_t)(message.len), g_uname,
-									clnet_info->realm, g_upwd)<1) {
+									clnet_info->realm, g_upwd, sht)<1) {
 							TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,"Wrong integrity in connect bind message received from server\n");
 							return -1;
 						}
@@ -1050,6 +1415,10 @@ static int turn_tcp_connection_bind(int verbose, app_ur_conn_info *clnet_info, a
 				} else if (stun_is_challenge_response_str(message.buf, (size_t)message.len,
 										&err_code,err_msg,sizeof(err_msg),
 										clnet_info->realm,clnet_info->nonce)) {
+					if(err_code == SHA_TOO_WEAK && (clnet_info->shatype == SHATYPE_SHA1)) {
+						clnet_info->shatype = SHATYPE_SHA256;
+						recalculate_restapi_hmac();
+					}
 					goto beg_cb;
 				} else if (stun_is_error_response(&message, &err_code,err_msg,sizeof(err_msg))) {
 					cb_received = 1;
@@ -1061,6 +1430,8 @@ static int turn_tcp_connection_bind(int verbose, app_ur_conn_info *clnet_info, a
 					/* Try again ? */
 				}
 			} else {
+				if(errorOK)
+					return 0;
 				perror("recv");
 				exit(-1);
 			}
@@ -1072,24 +1443,22 @@ static int turn_tcp_connection_bind(int verbose, app_ur_conn_info *clnet_info, a
 
 void tcp_data_connect(app_ur_session *elem, u32bits cid)
 {
-	int clnet_fd = socket(elem->pinfo.remote_addr.ss.ss_family, SOCK_STREAM, 0);
+	int clnet_fd = socket(elem->pinfo.remote_addr.ss.sa_family, SOCK_STREAM, 0);
 	if (clnet_fd < 0) {
 		perror("socket");
 		exit(-1);
 	}
 
-	socket_set_reusable(clnet_fd);
-
 	if (sock_bind_to_device(clnet_fd, client_ifname) < 0) {
 		TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,
 						"Cannot bind client socket to device %s\n", client_ifname);
 	}
-	set_sock_buf_size(clnet_fd, UR_CLIENT_SOCK_BUF_SIZE);
+	set_sock_buf_size(clnet_fd, (UR_CLIENT_SOCK_BUF_SIZE<<2));
 
 	++elem->pinfo.tcp_conn_number;
 	int i = (int)(elem->pinfo.tcp_conn_number-1);
 	elem->pinfo.tcp_conn=(app_tcp_conn_info**)realloc(elem->pinfo.tcp_conn,elem->pinfo.tcp_conn_number*sizeof(app_tcp_conn_info*));
-	elem->pinfo.tcp_conn[i]=(app_tcp_conn_info*)malloc(sizeof(app_tcp_conn_info));
+	elem->pinfo.tcp_conn[i]=(app_tcp_conn_info*)turn_malloc(sizeof(app_tcp_conn_info));
 	ns_bzero(elem->pinfo.tcp_conn[i],sizeof(app_tcp_conn_info));
 
 	elem->pinfo.tcp_conn[i]->tcp_data_fd = clnet_fd;
@@ -1099,7 +1468,7 @@ void tcp_data_connect(app_ur_session *elem, u32bits cid)
 
 	addr_set_port(&(elem->pinfo.tcp_conn[i]->tcp_data_local_addr),0);
 
-	addr_bind(clnet_fd, &(elem->pinfo.tcp_conn[i]->tcp_data_local_addr));
+	addr_bind(clnet_fd, &(elem->pinfo.tcp_conn[i]->tcp_data_local_addr), 1);
 
 	addr_get_from_sock(clnet_fd,&(elem->pinfo.tcp_conn[i]->tcp_data_local_addr));
 
@@ -1109,36 +1478,35 @@ void tcp_data_connect(app_ur_session *elem, u32bits cid)
 	    int err = 0;
 	    if (addr_connect(clnet_fd, &(elem->pinfo.remote_addr),&err) < 0) {
 	      if(err == EADDRINUSE) {
-		close(clnet_fd);
-		clnet_fd = socket(elem->pinfo.remote_addr.ss.ss_family, SOCK_STREAM, 0);
-		if (clnet_fd < 0) {
-		  perror("socket");
-		  exit(-1);
-		}
-		socket_set_reusable(clnet_fd);
-		if (sock_bind_to_device(clnet_fd, client_ifname) < 0) {
-		  TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,
-				"Cannot bind client socket to device %s\n", client_ifname);
-		}
-		set_sock_buf_size(clnet_fd, UR_CLIENT_SOCK_BUF_SIZE);
+	    	  socket_closesocket(clnet_fd);
+	    	  clnet_fd = socket(elem->pinfo.remote_addr.ss.sa_family, SOCK_STREAM, 0);
+	    	  if (clnet_fd < 0) {
+	    		  perror("socket");
+	    		  exit(-1);
+	    	  }
+	    	  if (sock_bind_to_device(clnet_fd, client_ifname) < 0) {
+	    		  TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,
+	    				  "Cannot bind client socket to device %s\n", client_ifname);
+	    	  }
+	    	  set_sock_buf_size(clnet_fd, UR_CLIENT_SOCK_BUF_SIZE<<2);
 
-		elem->pinfo.tcp_conn[i]->tcp_data_fd = clnet_fd;
+	    	  elem->pinfo.tcp_conn[i]->tcp_data_fd = clnet_fd;
 
-		addr_cpy(&(elem->pinfo.tcp_conn[i]->tcp_data_local_addr),&(elem->pinfo.local_addr));
+	    	  addr_cpy(&(elem->pinfo.tcp_conn[i]->tcp_data_local_addr),&(elem->pinfo.local_addr));
 
-		addr_set_port(&(elem->pinfo.tcp_conn[i]->tcp_data_local_addr),0);
+	    	  addr_set_port(&(elem->pinfo.tcp_conn[i]->tcp_data_local_addr),0);
 
-		addr_bind(clnet_fd, &(elem->pinfo.tcp_conn[i]->tcp_data_local_addr));
+	    	  addr_bind(clnet_fd, &(elem->pinfo.tcp_conn[i]->tcp_data_local_addr),1);
 
-		addr_get_from_sock(clnet_fd,&(elem->pinfo.tcp_conn[i]->tcp_data_local_addr));
+	    	  addr_get_from_sock(clnet_fd,&(elem->pinfo.tcp_conn[i]->tcp_data_local_addr));
 
-		continue;
+	    	  continue;
 
 	      } else {
-		perror("connect");
-		TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,
+	    	  perror("connect");
+	    	  TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,
 			      "%s: cannot connect to remote addr\n", __FUNCTION__);
-		exit(-1);
+	    	  exit(-1);
 	      }
 	    } else {
 	      break;
@@ -1155,12 +1523,12 @@ void tcp_data_connect(app_ur_session *elem, u32bits cid)
 		}
 	}
 
-	if(turn_tcp_connection_bind(clnet_verbose, &(elem->pinfo), elem->pinfo.tcp_conn[i])<0) {
+	if(turn_tcp_connection_bind(clnet_verbose, &(elem->pinfo), elem->pinfo.tcp_conn[i],0)<0) {
 		TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,
 				"%s: cannot BIND to tcp connection\n", __FUNCTION__);
 	} else {
 
-		evutil_make_socket_nonblocking(clnet_fd);
+		socket_set_nonblocking(clnet_fd);
 
 		struct event* ev = event_new(client_event_base,clnet_fd,
 					EV_READ|EV_PERSIST,client_input_handler,
