@@ -39,14 +39,17 @@
 
 #include <openssl/err.h>
 
-static const int verbose_packets=0;
+static int verbose_packets=0;
 
 static size_t current_clients_number = 0;
 
 static int start_full_timer=0;
 static u32bits tot_messages=0;
 static u32bits tot_send_messages=0;
+static u64bits tot_send_bytes = 0;
 static u32bits tot_recv_messages=0;
+static u64bits tot_recv_bytes = 0;
+static u64bits tot_send_dropped = 0;
 
 struct event_base* client_event_base=NULL;
 
@@ -58,12 +61,12 @@ static u64bits current_mstime = 0;
 
 static char buffer_to_send[65536]="\0";
 
-#define MAX_CLIENTS (1024)
 static int total_clients = 0;
 
-static app_ur_session* elems[MAX_CLIENTS];
+/* Patch for unlimited number of clients provided by ucudbm@gmail.com */
+static app_ur_session** elems = NULL;
 
-#define SLEEP_INTERVAL (2345)
+#define SLEEP_INTERVAL (234)
 
 int RTP_PACKET_INTERVAL = 20;
 
@@ -111,6 +114,7 @@ static app_ur_session* init_app_session(app_ur_session *ss) {
   if(ss) {
     ns_bzero(ss,sizeof(app_ur_session));
     ss->pinfo.fd=-1;
+    ss->pinfo.shatype = shatype;
   }
   return ss;
 }
@@ -118,7 +122,7 @@ static app_ur_session* init_app_session(app_ur_session *ss) {
 static app_ur_session* create_new_ss(void)
 {
 	++current_clients_number;
-	return init_app_session((app_ur_session*) malloc(sizeof(app_ur_session)));
+	return init_app_session((app_ur_session*) turn_malloc(sizeof(app_ur_session)));
 }
 
 static void uc_delete_session_elem_data(app_ur_session* cdi) {
@@ -139,16 +143,16 @@ static void uc_delete_session_elem_data(app_ur_session* cdi) {
 	      cdi->pinfo.tcp_conn[i]->tcp_data_ssl = NULL;
 	    }
 	    if(cdi->pinfo.tcp_conn[i]->tcp_data_fd>=0) {
-	      evutil_closesocket(cdi->pinfo.tcp_conn[i]->tcp_data_fd);
+	    	socket_closesocket(cdi->pinfo.tcp_conn[i]->tcp_data_fd);
 	      cdi->pinfo.tcp_conn[i]->tcp_data_fd=-1;
 	    }
-	    free(cdi->pinfo.tcp_conn[i]);
+	    turn_free(cdi->pinfo.tcp_conn[i], 111);
 	    cdi->pinfo.tcp_conn[i]=NULL;
 	  }
 	}
       }
       cdi->pinfo.tcp_conn_number=0;
-      free(cdi->pinfo.tcp_conn);
+      turn_free(cdi->pinfo.tcp_conn, 111);
       cdi->pinfo.tcp_conn=NULL;
     }
     if(cdi->pinfo.ssl && !(cdi->pinfo.broken)) {
@@ -162,7 +166,7 @@ static void uc_delete_session_elem_data(app_ur_session* cdi) {
 	    cdi->pinfo.ssl = NULL;
     }
     if(cdi->pinfo.fd>=0) {
-      evutil_closesocket(cdi->pinfo.fd);
+    	socket_closesocket(cdi->pinfo.fd);
     }
     cdi->pinfo.fd=-1;
   }
@@ -189,6 +193,17 @@ int send_buffer(app_ur_conn_info *clnet_info, stun_buffer* message, int data_con
 
 	char *buffer = (char*) (message->buf);
 
+	if(negative_protocol_test && (message->len>0)) {
+		if(random()%10 == 0) {
+			int np = (int)((unsigned long)random()%10);
+			while(np-->0) {
+				int pos = (int)((unsigned long)random()%(unsigned long)message->len);
+				int val = (int)((unsigned long)random()%256);
+				message->buf[pos]=(u08bits)val;
+			}
+		}
+	}
+
 	SSL *ssl = clnet_info->ssl;
 	ioa_socket_raw fd = clnet_info->fd;
 
@@ -214,10 +229,10 @@ int send_buffer(app_ur_conn_info *clnet_info, stun_buffer* message, int data_con
 
 			int len = 0;
 			do {
-				len = SSL_write(ssl, buffer, message->len);
-			} while (len < 0 && ((errno == EINTR) || (errno == ENOBUFS) || (errno == EAGAIN)));
+				len = SSL_write(ssl, buffer, (int)message->len);
+			} while (len < 0 && ((errno == EINTR) || (errno == ENOBUFS)));
 
-			if(len == message->len) {
+			if(len == (int)message->len) {
 				if (clnet_verbose) {
 					TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,
 							"buffer sent: size=%d\n",len);
@@ -267,12 +282,12 @@ int send_buffer(app_ur_conn_info *clnet_info, stun_buffer* message, int data_con
 		while (left > 0) {
 			do {
 				rc = send(fd, buffer, left, 0);
-			} while (rc < 0 && ((errno == EINTR) || (errno == ENOBUFS) || (errno
-							== EAGAIN)));
+			} while (rc < 0 && ((errno == EINTR) || (errno == ENOBUFS)));
 			if (rc > 0) {
 				left -= (size_t) rc;
 				buffer += rc;
 			} else {
+				tot_send_dropped+=1;
 				break;
 			}
 		}
@@ -286,26 +301,27 @@ int send_buffer(app_ur_conn_info *clnet_info, stun_buffer* message, int data_con
 	return ret;
 }
 
-int recv_buffer(app_ur_conn_info *clnet_info, stun_buffer* message, int sync, app_tcp_conn_info *atc) {
+int recv_buffer(app_ur_conn_info *clnet_info, stun_buffer* message, int sync,
+		app_tcp_conn_info *atc) {
 
 	int rc = 0;
 
 	ioa_socket_raw fd = clnet_info->fd;
-	if(atc)
+	if (atc)
 		fd = atc->tcp_data_fd;
 
 	SSL* ssl = clnet_info->ssl;
-	if(atc)
-		ssl=atc->tcp_data_ssl;
+	if (atc)
+		ssl = atc->tcp_data_ssl;
 
-	if(!use_secure && !use_tcp && fd>=0) {
+	if (!use_secure && !use_tcp && fd >= 0) {
 
 		/* Plain UDP */
 
 		do {
 			rc = recv(fd, message->buf, sizeof(message->buf) - 1, 0);
-			if(rc<0 && errno==EAGAIN && sync)
-				errno=EINTR;
+			if (rc < 0 && errno == EAGAIN && sync)
+				errno = EINTR;
 		} while (rc < 0 && (errno == EINTR));
 
 		if (rc < 0) {
@@ -314,32 +330,29 @@ int recv_buffer(app_ur_conn_info *clnet_info, stun_buffer* message, int sync, ap
 
 		message->len = rc;
 
-	} else if(use_secure && ssl && !(clnet_info->broken)) {
+	} else if (use_secure && ssl && !(clnet_info->broken)) {
 
 		/* TLS/DTLS */
 
 		int message_received = 0;
 		int cycle = 0;
-		while (!message_received && cycle++<100) {
+		while (!message_received && cycle++ < 100) {
 
 			if (SSL_get_shutdown(ssl))
 				return -1;
 
 			rc = 0;
 			do {
-				rc = SSL_read(ssl, message->buf,
-								sizeof(message->buf) - 1);
-				if(rc<0 && errno==EAGAIN && sync)
+				rc = SSL_read(ssl, message->buf, sizeof(message->buf) - 1);
+				if (rc < 0 && errno == EAGAIN && sync)
 					continue;
 			} while (rc < 0 && (errno == EINTR));
 
 			if (rc > 0) {
 
 				if (clnet_verbose) {
-					TURN_LOG_FUNC(
-									TURN_LOG_LEVEL_INFO,
-									"response received: size=%d\n",
-									rc);
+					TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,
+							"response received: size=%d\n", rc);
 				}
 				message->len = rc;
 				message_received = 1;
@@ -348,7 +361,7 @@ int recv_buffer(app_ur_conn_info *clnet_info, stun_buffer* message, int sync, ap
 
 				int sslerr = SSL_get_error(ssl, rc);
 
-				switch (sslerr){
+				switch (sslerr) {
 				case SSL_ERROR_NONE:
 					/* Try again ? */
 					break;
@@ -363,96 +376,92 @@ int recv_buffer(app_ur_conn_info *clnet_info, stun_buffer* message, int sync, ap
 					break;
 				case SSL_ERROR_SYSCALL:
 					TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,
-									"Socket read error 111.999: \n");
+							"Socket read error 111.999: \n");
 					if (handle_socket_error())
 						break;
-				case SSL_ERROR_SSL:
-				{
-					TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,
-									"SSL write error: \n");
+				case SSL_ERROR_SSL: {
+					TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "SSL write error: \n");
 					char buf[1024];
-					TURN_LOG_FUNC(
-									TURN_LOG_LEVEL_INFO,
-									"%s (%d)\n",
-									ERR_error_string(
-													ERR_get_error(),
-													buf),
-									SSL_get_error(
-													ssl,
-													rc));
+					TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "%s (%d)\n",
+							ERR_error_string(ERR_get_error(), buf),
+							SSL_get_error(ssl, rc));
 				}
 				default:
 					clnet_info->broken = 1;
 					TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,
-									"Unexpected error while reading: rc=%d, sslerr=%d\n",rc,sslerr);
+							"Unexpected error while reading: rc=%d, sslerr=%d\n",
+							rc, sslerr);
 					return -1;
 				}
 
-				if(!sync)
-				  break;
+				if (!sync)
+					break;
 			}
 		}
 
-	} else if(!use_secure && use_tcp && fd>=0){
+	} else if (!use_secure && use_tcp && fd >= 0) {
 
-	  /* Plain TCP */
+		/* Plain TCP */
 
-	  do {
-	    rc = recv(fd, message->buf, sizeof(message->buf) - 1, MSG_PEEK);
-	    if((rc<0) && (errno==EAGAIN) && sync) {
-	      errno=EINTR;
-	    }
-	  } while (rc < 0 && (errno == EINTR));
-	    
-	  if(rc>0) {
-	    int mlen = rc;
-	    size_t app_msg_len = (size_t)rc;
-	    if(!atc) {
-	      mlen = stun_get_message_len_str(message->buf, rc, 1, &app_msg_len);
-	    } else {
-	      if(!sync)
-		mlen = clmessage_length;
-
-	      if(mlen>clmessage_length)
-		mlen = clmessage_length;
-
-	      app_msg_len = (size_t)mlen;
-	    }
-	      
-	    if(mlen>0) {
-
-	      int rcr = 0;
-	      int rsf = 0;
-	      int cycle = 0;
-	      while(rsf<mlen && cycle++<128) {
 		do {
-		  rcr = recv(fd, message->buf+rsf, (size_t)mlen-(size_t)rsf, 0);
-		  if(rcr<0 && errno==EAGAIN && sync)
-		    errno=EINTR;
-		} while (rcr < 0 && (errno == EINTR));
-	      
-		if (rcr > 0)
-		  rsf+= rcr;
-	      
-	      }
+			rc = recv(fd, message->buf, sizeof(message->buf) - 1, MSG_PEEK);
+			if ((rc < 0) && (errno == EAGAIN) && sync) {
+				errno = EINTR;
+			}
+		} while (rc < 0 && (errno == EINTR));
 
-	      if(rsf<1)
-		return -1;
+		if (rc > 0) {
+			int mlen = rc;
+			size_t app_msg_len = (size_t) rc;
+			if (!atc) {
+				mlen = stun_get_message_len_str(message->buf, rc, 1,
+						&app_msg_len);
+			} else {
+				if (!sync)
+					mlen = clmessage_length;
 
-	      if(rsf<(int)app_msg_len) {
-		      if((size_t)(app_msg_len/(size_t)rsf)*((size_t)(rsf)) != app_msg_len) {
-			      return -1;
-		      }
-	      }
+				if (mlen > clmessage_length)
+					mlen = clmessage_length;
 
-	      message->len = app_msg_len;
+				app_msg_len = (size_t) mlen;
+			}
 
-	      rc = app_msg_len;
+			if (mlen > 0) {
 
-	    } else {
-	      rc = 0;
-	    }
-	  }
+				int rcr = 0;
+				int rsf = 0;
+				int cycle = 0;
+				while (rsf < mlen && cycle++ < 128) {
+					do {
+						rcr = recv(fd, message->buf + rsf,
+								(size_t) mlen - (size_t) rsf, 0);
+						if (rcr < 0 && errno == EAGAIN && sync)
+							errno = EINTR;
+					} while (rcr < 0 && (errno == EINTR));
+
+					if (rcr > 0)
+						rsf += rcr;
+
+				}
+
+				if (rsf < 1)
+					return -1;
+
+				if (rsf < (int) app_msg_len) {
+					if ((size_t) (app_msg_len / (size_t) rsf) * ((size_t) (rsf))
+							!= app_msg_len) {
+						return -1;
+					}
+				}
+
+				message->len = app_msg_len;
+
+				rc = app_msg_len;
+
+			} else {
+				rc = 0;
+			}
+		}
 	}
 
 	return rc;
@@ -472,6 +481,7 @@ static int client_read(app_ur_session *elem, int is_tcp_data, app_tcp_conn_info
 	int err_code = 0;
 	u08bits err_msg[129];
 	int rc = 0;
+	int applen = 0;
 
 	if (clnet_verbose && verbose_packets) {
 		TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "before read ...\n");
@@ -493,32 +503,17 @@ static int client_read(app_ur_session *elem, int is_tcp_data, app_tcp_conn_info
 
 		size_t buffers = 1;
 
-		 if(is_tcp_data) {
-			if (elem->in_buffer.len >= 0) {
-				if (elem->in_buffer.len > clmessage_length) {
-					if(clnet_verbose) {
-						TURN_LOG_FUNC(
-						TURN_LOG_LEVEL_INFO,
-						"WARNING: received multiple buffers: length: %d, length of single buffer must be %d\n",
-						rc, clmessage_length);
-					}
-					buffers=elem->in_buffer.len/clmessage_length;
-				} else if (elem->in_buffer.len < clmessage_length) {
-					TURN_LOG_FUNC(
-						TURN_LOG_LEVEL_INFO,
-						"ERROR: received wrong buffer size: length: %d, must be %d; len=%d\n",
-						rc, clmessage_length, (int)elem->in_buffer.len);
-					return rc;
-				}
-
-				mi = (message_info*)(elem->in_buffer.buf);
-			}
+		if(is_tcp_data) {
+		   if ((int)elem->in_buffer.len == clmessage_length) {
+		     mi = (message_info*)(elem->in_buffer.buf);
+		   }
 		} else if (stun_is_indication(&(elem->in_buffer))) {
 
 			if(use_short_term) {
+				SHATYPE sht = elem->pinfo.shatype;
 				if(stun_check_message_integrity_str(get_turn_credentials_type(),
 							elem->in_buffer.buf, (size_t)(elem->in_buffer.len), g_uname,
-							elem->pinfo.realm, g_upwd)<1) {
+							elem->pinfo.realm, g_upwd, sht)<1) {
 					TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,"Wrong integrity in indication message 0x%x received from server\n",(unsigned int)stun_get_method(&(elem->in_buffer)));
 					return -1;
 				}
@@ -528,11 +523,22 @@ static int client_read(app_ur_session *elem, int is_tcp_data, app_tcp_conn_info
 
 			if((method == STUN_METHOD_CONNECTION_ATTEMPT)&& is_TCP_relay()) {
 			  stun_attr_ref sar = stun_attr_get_first(&(elem->in_buffer));
-			  if(sar) {
-			    u32bits cid = *((const u32bits*)stun_attr_get_value(sar));
-			    tcp_data_connect(elem,cid);
-			    return rc;
+			  u32bits cid = 0;
+			  while(sar) {
+				  int attr_type = stun_attr_get_type(sar);
+				  if(attr_type == STUN_ATTRIBUTE_CONNECTION_ID) {
+					  cid = *((const u32bits*)stun_attr_get_value(sar));
+					  break;
+				  }
+				  sar = stun_attr_get_next_str(elem->in_buffer.buf,elem->in_buffer.len,sar);
 			  }
+			  if(negative_test) {
+				  tcp_data_connect(elem,(u64bits)random());
+			  } else {
+				  /* positive test */
+				  tcp_data_connect(elem,cid);
+			  }
+			  return rc;
 			} else if (method != STUN_METHOD_DATA) {
 				TURN_LOG_FUNC(
 						TURN_LOG_LEVEL_INFO,
@@ -548,8 +554,10 @@ static int client_read(app_ur_session *elem, int is_tcp_data, app_tcp_conn_info
 				}
 
 				int rlen = stun_attr_get_len(sar);
+				applen = rlen;
 				if (rlen != clmessage_length) {
 					TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "ERROR: received DATA message has wrong len: %d, must be %d\n", rlen, clmessage_length);
+					tot_recv_bytes += applen;
 					return rc;
 				}
 
@@ -561,9 +569,10 @@ static int client_read(app_ur_session *elem, int is_tcp_data, app_tcp_conn_info
 		} else if (stun_is_success_response(&(elem->in_buffer))) {
 
 			if(elem->pinfo.nonce[0] || use_short_term) {
+				SHATYPE sht = elem->pinfo.shatype;
 				if(stun_check_message_integrity_str(get_turn_credentials_type(),
 								elem->in_buffer.buf, (size_t)(elem->in_buffer.len), g_uname,
-								elem->pinfo.realm, g_upwd)<0) {
+								elem->pinfo.realm, g_upwd, sht)<1) {
 					TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,"Wrong integrity in success message 0x%x received from server\n",(unsigned int)stun_get_method(&(elem->in_buffer)));
 					return -1;
 				}
@@ -571,16 +580,27 @@ static int client_read(app_ur_session *elem, int is_tcp_data, app_tcp_conn_info
 
 			if(is_TCP_relay() && (stun_get_method(&(elem->in_buffer)) == STUN_METHOD_CONNECT)) {
 				stun_attr_ref sar = stun_attr_get_first(&(elem->in_buffer));
-				if(sar) {
-					u32bits cid = *((const u32bits*)stun_attr_get_value(sar));
-					tcp_data_connect(elem,cid);
+				u32bits cid = 0;
+				while(sar) {
+				  int attr_type = stun_attr_get_type(sar);
+				  if(attr_type == STUN_ATTRIBUTE_CONNECTION_ID) {
+					  cid = *((const u32bits*)stun_attr_get_value(sar));
+					  break;
+				  }
+				  sar = stun_attr_get_next_str(elem->in_buffer.buf,elem->in_buffer.len,sar);
 				}
+				tcp_data_connect(elem,cid);
 			}
 
 			return rc;
 		} else if (stun_is_challenge_response_str(elem->in_buffer.buf, (size_t)elem->in_buffer.len,
 							&err_code,err_msg,sizeof(err_msg),
 							clnet_info->realm,clnet_info->nonce)) {
+			if(err_code == SHA_TOO_WEAK && (elem->pinfo.shatype == SHATYPE_SHA1)) {
+				elem->pinfo.shatype = SHATYPE_SHA256;
+				recalculate_restapi_hmac();
+			}
+
 			if(is_TCP_relay() && (stun_get_method(&(elem->in_buffer)) == STUN_METHOD_CONNECT)) {
 				turn_tcp_connect(clnet_verbose, &(elem->pinfo), &(elem->pinfo.peer_addr));
 			} else if(stun_get_method(&(elem->in_buffer)) == STUN_METHOD_REFRESH) {
@@ -598,8 +618,8 @@ static int client_read(app_ur_session *elem, int is_tcp_data, app_tcp_conn_info
 			}
 
 			if (elem->in_buffer.len >= 4) {
-				if (((elem->in_buffer.len-4) < clmessage_length) ||
-					((elem->in_buffer.len-4) > clmessage_length + 3)) {
+				if (((int)(elem->in_buffer.len-4) < clmessage_length) ||
+					((int)(elem->in_buffer.len-4) > clmessage_length + 3)) {
 					TURN_LOG_FUNC(
 							TURN_LOG_LEVEL_INFO,
 							"ERROR: received buffer have wrong length: %d, must be %d, len=%d\n",
@@ -608,6 +628,7 @@ static int client_read(app_ur_session *elem, int is_tcp_data, app_tcp_conn_info
 				}
 
 				mi = (message_info*)(elem->in_buffer.buf + 4);
+				applen = elem->in_buffer.len -4;
 			}
 		} else {
 			TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,
@@ -646,6 +667,10 @@ static int client_read(app_ur_session *elem, int is_tcp_data, app_tcp_conn_info
 
 		elem->rmsgnum+=buffers;
 		tot_recv_messages+=buffers;
+		if(applen > 0)
+			tot_recv_bytes += applen;
+		else
+			tot_recv_bytes += elem->in_buffer.len;
 		elem->recvtimems=current_mstime;
 		elem->wait_cycles = 0;
 
@@ -697,6 +722,7 @@ static int client_write(app_ur_session *elem) {
 			  ++elem->wmsgnum;
 			  elem->to_send_timems += RTP_PACKET_INTERVAL;
 			  tot_send_messages++;
+			  tot_send_bytes += clmessage_length;
 		  }
 		  return 0;
 	  }
@@ -745,6 +771,7 @@ static int client_write(app_ur_session *elem) {
 				TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "wrote %d bytes\n", (int) rc);
 	  }
       tot_send_messages++;
+      tot_send_bytes += clmessage_length;
     } else {
     	return -1;
     }
@@ -824,6 +851,7 @@ static int start_client(const char *remote_address, int port,
   app_ur_conn_info clnet_info_probe; /* for load balancing probe */
   ns_bzero(&clnet_info_probe,sizeof(clnet_info_probe));
   clnet_info_probe.fd = -1;
+  clnet_info_probe.shatype = shatype;
 
   app_ur_conn_info *clnet_info=&(ss->pinfo);
   app_ur_conn_info *clnet_info_rtcp=NULL;
@@ -846,14 +874,14 @@ static int start_client(const char *remote_address, int port,
   	clnet_info_probe.ssl = NULL;
   	clnet_info_probe.fd = -1;
   } else if(clnet_info_probe.fd != -1) {
-  	close(clnet_info_probe.fd);
-  	clnet_info_probe.fd = -1;
+	  socket_closesocket(clnet_info_probe.fd);
+	  clnet_info_probe.fd = -1;
   }
   
-  evutil_make_socket_nonblocking(clnet_info->fd);
+  socket_set_nonblocking(clnet_info->fd);
   
   if(!no_rtcp) 
-    evutil_make_socket_nonblocking(clnet_info_rtcp->fd);
+	  socket_set_nonblocking(clnet_info_rtcp->fd);
   
   struct event* ev = event_new(client_event_base,clnet_info->fd,
 				EV_READ|EV_PERSIST,client_input_handler,
@@ -919,6 +947,7 @@ static int start_c2c(const char *remote_address, int port,
   app_ur_conn_info clnet_info_probe; /* for load balancing probe */
   ns_bzero(&clnet_info_probe,sizeof(clnet_info_probe));
   clnet_info_probe.fd = -1;
+  clnet_info_probe.shatype = shatype;
 
   app_ur_conn_info *clnet_info1=&(ss1->pinfo);
   app_ur_conn_info *clnet_info1_rtcp=NULL;
@@ -951,19 +980,19 @@ static int start_c2c(const char *remote_address, int port,
 	clnet_info_probe.ssl = NULL;
 	clnet_info_probe.fd = -1;
   } else if(clnet_info_probe.fd != -1) {
-	close(clnet_info_probe.fd);
-	clnet_info_probe.fd = -1;
+	  socket_closesocket(clnet_info_probe.fd);
+	  clnet_info_probe.fd = -1;
   }
   
-  evutil_make_socket_nonblocking(clnet_info1->fd);
+  socket_set_nonblocking(clnet_info1->fd);
   
   if(!no_rtcp)
-    evutil_make_socket_nonblocking(clnet_info1_rtcp->fd);
+	  socket_set_nonblocking(clnet_info1_rtcp->fd);
   
-  evutil_make_socket_nonblocking(clnet_info2->fd);
+  socket_set_nonblocking(clnet_info2->fd);
   
   if(!no_rtcp)
-    evutil_make_socket_nonblocking(clnet_info2_rtcp->fd);
+	  socket_set_nonblocking(clnet_info2_rtcp->fd);
   
   struct event* ev1 = event_new(client_event_base,clnet_info1->fd,
 				EV_READ|EV_PERSIST,client_input_handler,
@@ -1062,13 +1091,15 @@ static int refresh_channel(app_ur_session* elem, u16bits method)
 
 	if (!addr_any(&(elem->pinfo.peer_addr))) {
 
-		if (!method || (method == STUN_METHOD_CREATE_PERMISSION)) {
-			stun_init_request(STUN_METHOD_CREATE_PERMISSION, &message);
-			stun_attr_add_addr(&message, STUN_ATTRIBUTE_XOR_PEER_ADDRESS, &(elem->pinfo.peer_addr));
-			if(add_integrity(clnet_info, &message)<0) return -1;
-			if(use_fingerprints)
+		if(!no_permissions) {
+			if (!method || (method == STUN_METHOD_CREATE_PERMISSION)) {
+				stun_init_request(STUN_METHOD_CREATE_PERMISSION, &message);
+				stun_attr_add_addr(&message, STUN_ATTRIBUTE_XOR_PEER_ADDRESS, &(elem->pinfo.peer_addr));
+				if(add_integrity(clnet_info, &message)<0) return -1;
+				if(use_fingerprints)
 				    stun_attr_add_fingerprint_str(message.buf, (size_t*) &(message.len));
-			send_buffer(&(elem->pinfo), &message, 0,0);
+				send_buffer(&(elem->pinfo), &message, 0,0);
+			}
 		}
 
 		if (!method || (method == STUN_METHOD_CHANNEL_BIND)) {
@@ -1176,12 +1207,14 @@ void start_mclient(const char *remote_address, int port,
 	      ++mclient;
 	}
 
+	elems = (app_ur_session**)malloc(sizeof(app_ur_session)*((mclient*2)+1)+sizeof(void*));
+
 	__turn_getMSTime();
 	u32bits stime = current_time;
 
-	memset(buffer_to_send, random(), clmessage_length);
+	memset(buffer_to_send, 7, clmessage_length);
 
-	client_event_base = event_base_new();
+	client_event_base = turn_event_base_new();
 
 	int i = 0;
 	int tot_clients = 0;
@@ -1189,7 +1222,7 @@ void start_mclient(const char *remote_address, int port,
 	if(c2c) {
 	  if(!no_rtcp)
 	    for (i = 0; i < (mclient >> 2); i++) {
-	      usleep(SLEEP_INTERVAL);
+	      if(!dos) usleep(SLEEP_INTERVAL);
 	      if (start_c2c(remote_address, port, ifname, local_address,
 			    messagenumber, i << 2) < 0) {
 	    	  exit(-1);
@@ -1198,7 +1231,7 @@ void start_mclient(const char *remote_address, int port,
 	    }
 	  else
 	    for (i = 0; i < (mclient >> 1); i++) {
-	      usleep(SLEEP_INTERVAL);
+	      if(!dos) usleep(SLEEP_INTERVAL);
 	      if (start_c2c(remote_address, port, ifname, local_address,
 			    messagenumber, i << 1) < 0) {
 	    	  exit(-1);
@@ -1208,7 +1241,7 @@ void start_mclient(const char *remote_address, int port,
 	} else {
 	  if(!no_rtcp)
 	    for (i = 0; i < (mclient >> 1); i++) {
-	      usleep(SLEEP_INTERVAL);
+	      if(!dos) usleep(SLEEP_INTERVAL);
 	      if (start_client(remote_address, port, ifname, local_address,
 			       messagenumber, i << 1) < 0) {
 	    	  exit(-1);
@@ -1217,7 +1250,7 @@ void start_mclient(const char *remote_address, int port,
 	    }
 	  else 
 	    for (i = 0; i < mclient; i++) {
-	      usleep(SLEEP_INTERVAL);
+	      if(!dos) usleep(SLEEP_INTERVAL);
 	      if (start_client(remote_address, port, ifname, local_address,
 			       messagenumber, i) < 0) {
 	    	  exit(-1);
@@ -1226,6 +1259,9 @@ void start_mclient(const char *remote_address, int port,
 	    }
 	}
 
+	if(dos)
+		_exit(0);
+
 	total_clients = tot_clients;
 
 	__turn_getMSTime();
@@ -1246,14 +1282,13 @@ void start_mclient(const char *remote_address, int port,
 					int connect_err = 0;
 					socket_connect(elems[i]->pinfo.fd, &(elems[i]->pinfo.remote_addr), &connect_err);
 				}
-			} else {
+			} else if((i%2) == 0) {
 				if (turn_tcp_connect(clnet_verbose, &(elems[i]->pinfo), &(elems[i]->pinfo.peer_addr)) < 0) {
 					exit(-1);
 				}
 			}
 		}
 		run_events(1);
-		elems[i]->to_send_timems = current_mstime + 1000 + ((u32bits)random())%500;
 	}
 
 	__turn_getMSTime();
@@ -1272,42 +1307,46 @@ void start_mclient(const char *remote_address, int port,
 				for(i=0;i<total_clients;++i) {
 					if(elems[i]->pinfo.is_peer) {
 						completed+=1;
-					} else if(elems[i]->pinfo.tcp_conn_number==1 &&
+					} else if(elems[i]->pinfo.tcp_conn_number>0 &&
 							elems[i]->pinfo.tcp_conn[0]->tcp_data_bound) {
 						completed += elems[i]->pinfo.tcp_conn_number;
 					}
-					elems[i]->to_send_timems = current_mstime + ((u32bits)random())%1500;
 				}
-				if((completed) == total_clients)
+				if(completed >= total_clients)
 					break;
 			} else {
 				for(i=0;i<total_clients;++i) {
-					if(elems[i]->pinfo.tcp_conn_number==2 &&
-						elems[i]->pinfo.tcp_conn[0]->tcp_data_bound &&
-						elems[i]->pinfo.tcp_conn[1]->tcp_data_bound) {
+					if(elems[i]->pinfo.tcp_conn_number>0 &&
+							elems[i]->pinfo.tcp_conn[0]->tcp_data_bound) {
 						completed += elems[i]->pinfo.tcp_conn_number;
 					}
-					elems[i]->to_send_timems = current_mstime + ((u32bits)random())%1500;
 				}
-				if((completed>>1) == total_clients)
+				if(completed >= total_clients)
 					break;
 			}
 			run_events(0);
 			if(current_time > connect_wait_start_time + STARTING_TCP_RELAY_TIME) {
 				TURN_LOG_FUNC(TURN_LOG_LEVEL_WARNING, "WARNING: %d connections are not completed\n",
-						(int)((total_clients<<1) - completed));
+						(int)(total_clients - completed));
 				break;
 			}
 		}
 	}
 
+	__turn_getMSTime();
+	stime = current_time;
+
+	for(i=0;i<total_clients;i++) {
+		elems[i]->to_send_timems = current_mstime + 1000 + ((u32bits)random())%5000;
+	}
+
 	tot_messages = elems[0]->tot_msgnum * total_clients;
 
 	start_full_timer = 1;
 
 	while (1) {
 
-		run_events(0);
+		run_events(1);
 
 		int msz = (int)current_clients_number;
 		if (msz < 1) {
@@ -1316,19 +1355,27 @@ void start_mclient(const char *remote_address, int port,
 
 		if(show_statistics) {
 			TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,
-				      "%s: msz=%d, tot_send_messages=%lu, tot_recv_messages=%lu\n",
+				      "%s: msz=%d, tot_send_msgs=%lu, tot_recv_msgs=%lu, tot_send_bytes ~ %llu, tot_recv_bytes ~ %llu\n",
 				      __FUNCTION__, msz, (unsigned long) tot_send_messages,
-				      (unsigned long) tot_recv_messages);
+				      (unsigned long) tot_recv_messages, 
+				      (unsigned long long) tot_send_bytes,
+				      (unsigned long long) tot_recv_bytes);
 			show_statistics=0;
 		}
 	}
 
 	TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,
-		      "%s: tot_send_messages=%lu, tot_recv_messages=%lu\n", 
+		      "%s: tot_send_msgs=%lu, tot_recv_msgs=%lu\n",
 		      __FUNCTION__,
 		      (unsigned long) tot_send_messages,
 		      (unsigned long) tot_recv_messages);
 
+	TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO,
+		      "%s: tot_send_bytes ~ %lu, tot_recv_bytes ~ %lu\n",
+		      __FUNCTION__,
+		      (unsigned long) tot_send_bytes,
+		      (unsigned long) tot_recv_bytes);
+
 	if (client_event_base)
 		event_base_free(client_event_base);
 
@@ -1339,8 +1386,9 @@ void start_mclient(const char *remote_address, int port,
 
 	TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "Total transmit time is %u\n",
 			((unsigned int)(current_time - stime)));
-	TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "Total lost packets %llu (%f%c)\n",
-				(unsigned long long)total_loss, (((double)total_loss/(double)tot_send_messages)*100.00),'%');
+	TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "Total lost packets %llu (%f%c), total send dropped %llu (%f%c)\n",
+				(unsigned long long)total_loss, (((double)total_loss/(double)tot_send_messages)*100.00),'%',
+				(unsigned long long)tot_send_dropped, (((double)tot_send_dropped/(double)(tot_send_messages+tot_send_dropped))*100.00),'%');
 	TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO, "Average round trip delay %f ms; min = %lu ms, max = %lu ms\n",
 				((double)total_latency/(double)((tot_recv_messages<1) ? 1 : tot_recv_messages)),
 				(unsigned long)min_latency,
@@ -1349,6 +1397,8 @@ void start_mclient(const char *remote_address, int port,
 				((double)total_jitter/(double)tot_recv_messages),
 				(unsigned long)min_jitter,
 				(unsigned long)max_jitter);
+
+	free(elems);
 }
 
 ///////////////////////////////////////////
@@ -1363,13 +1413,13 @@ turn_credential_type get_turn_credentials_type(void)
 int add_integrity(app_ur_conn_info *clnet_info, stun_buffer *message)
 {
 	if(use_short_term) {
-		if(stun_attr_add_integrity_by_user_short_term_str(message->buf, (size_t*)&(message->len), g_uname, g_upwd)<0) {
+		if(stun_attr_add_integrity_by_user_short_term_str(message->buf, (size_t*)&(message->len), g_uname, g_upwd, clnet_info->shatype)<0) {
 			TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO," Cannot add integrity to the message\n");
 			return -1;
 		}
 	} else if(clnet_info->nonce[0]) {
 		if(stun_attr_add_integrity_by_user_str(message->buf, (size_t*)&(message->len), g_uname,
-					clnet_info->realm, g_upwd, clnet_info->nonce)<0) {
+					clnet_info->realm, g_upwd, clnet_info->nonce, clnet_info->shatype)<0) {
 			TURN_LOG_FUNC(TURN_LOG_LEVEL_INFO," Cannot add integrity to the message\n");
 			return -1;
 		}
@@ -1379,3 +1429,4 @@ int add_integrity(app_ur_conn_info *clnet_info, stun_buffer *message)
 }
 
 ///////////////////////////////////////////
+
