@@ -31,30 +31,66 @@
 #include "ns_turn_msg.h"
 #include "ns_turn_msg_addr.h"
 
-#if defined(__USE_OPENSSL__)
-
 ///////////// Security functions implementation from ns_turn_msg.h ///////////
 
 #include <openssl/md5.h>
 #include <openssl/hmac.h>
 #include <openssl/ssl.h>
+#include <openssl/err.h>
+#include <openssl/rand.h>
 
-int stun_calculate_hmac(u08bits *buf, size_t len, u08bits *key, size_t keylen, u08bits *hmac)
+#include <stdlib.h>
+
+///////////
+
+long turn_random(void)
 {
-	if (!HMAC(EVP_sha1(), key, keylen, buf, len, hmac, NULL)) {
-		return -1;
-	} else {
-		return 0;
+	long ret = 0;
+	if(!RAND_bytes((unsigned char *)&ret,sizeof(ret)))
+		ret = random();
+	return ret;
+}
+
+void turn_random32_size(u32bits *ar, size_t sz)
+{
+	if(!RAND_bytes((unsigned char *)ar, sz<<2)<0) {
+		size_t i;
+		for(i=0;i<sz;++i) {
+			ar[i] = (u32bits)random();
+		}
 	}
 }
 
-int stun_produce_integrity_key_str(u08bits *uname, u08bits *realm, u08bits *upwd, hmackey_t key)
+int stun_calculate_hmac(const u08bits *buf, size_t len, const u08bits *key, size_t keylen, u08bits *hmac, unsigned int *hmac_len, SHATYPE shatype)
 {
-	MD5_CTX ctx;
+	ERR_clear_error();
+	UNUSED_ARG(shatype);
+
+#if !defined(OPENSSL_NO_SHA256) && defined(SSL_TXT_SHA256)
+	if(shatype == SHATYPE_SHA256) {
+	  if (!HMAC(EVP_sha256(), key, keylen, buf, len, hmac, hmac_len)) {
+	    return -1;
+	  }
+	} else
+#endif
+
+	  if (!HMAC(EVP_sha1(), key, keylen, buf, len, hmac, hmac_len)) {
+	    return -1;
+	  }
+
+	return 0;
+}
+
+int stun_produce_integrity_key_str(u08bits *uname, u08bits *realm, u08bits *upwd, hmackey_t key, SHATYPE shatype)
+{
+	ERR_clear_error();
+	UNUSED_ARG(shatype);
+
 	size_t ulen = strlen((s08bits*)uname);
 	size_t rlen = strlen((s08bits*)realm);
 	size_t plen = strlen((s08bits*)upwd);
 	size_t sz = ulen+1+rlen+1+plen+1+10;
+	size_t strl = ulen+1+rlen+1+plen;
 	u08bits *str = (u08bits*)malloc(sz+1);
 
 	strncpy((s08bits*)str,(s08bits*)uname,sz);
@@ -62,18 +98,29 @@ int stun_produce_integrity_key_str(u08bits *uname, u08bits *realm, u08bits *upwd
 	strncpy((s08bits*)str+ulen+1,(s08bits*)realm,sz-ulen-1);
 	str[ulen+1+rlen]=':';
 	strncpy((s08bits*)str+ulen+1+rlen+1,(s08bits*)upwd,sz-ulen-1-rlen-1);
-	str[sz]=0;
+	str[strl]=0;
+
+#if !defined(OPENSSL_NO_SHA256) && defined(SSL_TXT_SHA256)
+	if(shatype == SHATYPE_SHA256) {
+		unsigned int keylen = 0;
+		EVP_MD_CTX ctx;
+		EVP_DigestInit(&ctx,EVP_sha256());
+		EVP_DigestUpdate(&ctx,str,strl);
+		EVP_DigestFinal(&ctx,key,&keylen);
+	} else
+#endif
+	{
+		MD5_CTX ctx;
+		MD5_Init(&ctx);
+		MD5_Update(&ctx,str,strl);
+		MD5_Final(key,&ctx);
+	}
 
-	MD5_Init(&ctx);
-	MD5_Update(&ctx,str,ulen+1+rlen+1+plen);
-	MD5_Final(key,&ctx);
 	free(str);
 
 	return 0;
 }
 
-#endif
-
 /////////////////////////////////////////////////////////////////
 
 static u32bits ns_crc32(const u08bits *buffer, u32bits len);
@@ -148,11 +195,14 @@ int old_stun_is_command_message_str(const u08bits* buf, size_t blen, u32bits *co
 	if (buf && blen >= STUN_HEADER_LENGTH) {
 		if (!STUN_VALID_CHANNEL(nswap16(((const u16bits*)buf)[0]))) {
 			if ((((u08bits) buf[0]) & ((u08bits) (0xC0))) == 0) {
-				u16bits len = nswap16(((const u16bits*)(buf))[1]);
-				if ((len & 0x0003) == 0) {
-					if ((size_t) (len + STUN_HEADER_LENGTH) == blen) {
-						*cookie = nswap32(((const u32bits*)(buf))[1]);
-						return 1;
+				if (nswap32(((const u32bits*)(buf))[1])
+						!= STUN_MAGIC_COOKIE) {
+					u16bits len = nswap16(((const u16bits*)(buf))[1]);
+					if ((len & 0x0003) == 0) {
+						if ((size_t) (len + STUN_HEADER_LENGTH) == blen) {
+							*cookie = nswap32(((const u32bits*)(buf))[1]);
+							return 1;
+						}
 					}
 				}
 			}
@@ -231,7 +281,7 @@ int stun_is_challenge_response_str(const u08bits* buf, size_t len, int *err_code
 {
 	int ret = stun_is_error_response_str(buf, len, err_code, err_msg, err_msg_size);
 
-	if(ret && (*err_code == 401 || *err_code == 438)) {
+	if(ret && (((*err_code) == 401) || ((*err_code) == 438) || ((*err_code) == SHA_TOO_WEAK))) {
 
 		stun_attr_ref sar = stun_attr_get_first_by_type_str(buf,len,STUN_ATTRIBUTE_REALM);
 		if(sar) {
@@ -367,7 +417,7 @@ static void stun_init_error_response_common_str(u08bits* buf, size_t *len,
 		};
 	}
 
-	u08bits avalue[129];
+	u08bits avalue[513];
 	avalue[0] = 0;
 	avalue[1] = 0;
 	avalue[2] = (u08bits) (error_code / 100);
@@ -376,6 +426,14 @@ static void stun_init_error_response_common_str(u08bits* buf, size_t *len,
 	avalue[sizeof(avalue)-1]=0;
 	int alen = 4 + strlen((const s08bits*) (avalue+4));
 
+	//"Manual" padding for compatibility with classic old stun:
+	{
+		int rem = alen % 4;
+		if(rem) {
+			alen +=(4-rem);
+		}
+	}
+
 	stun_attr_add_str(buf, len, STUN_ATTRIBUTE_ERROR_CODE, (u08bits*) avalue, alen);
 	if (id) {
 		stun_tid_message_cpy(buf, id);
@@ -443,6 +501,9 @@ int stun_is_channel_message_str(const u08bits *buf, size_t *blen, u16bits* chnum
 	datalen_header = ((const u16bits*)buf)[1];
 	datalen_header = nswap16(datalen_header);
 
+	if (datalen_header > datalen_actual)
+		return 0;
+
 	if (datalen_header != datalen_actual) {
 
 		/* maybe there are padding bytes for 32-bit alignment. Mandatory for TCP. Optional for UDP */
@@ -471,6 +532,62 @@ int stun_is_channel_message_str(const u08bits *buf, size_t *blen, u16bits* chnum
 
 ////////// STUN message ///////////////////////////////
 
+static inline int sheadof(const char *head, const char* full)
+{
+	while(*head) {
+		if(*head != *full)
+			return 0;
+		++head;++full;
+	}
+	return 1;
+}
+
+static inline const char* findstr(const char *hay, size_t slen, const char *needle)
+{
+	const char *ret = NULL;
+
+	if(hay && slen && needle) {
+		size_t nlen=strlen(needle);
+		if(nlen<=slen) {
+			size_t smax = slen-nlen+1;
+			size_t i;
+			const char *sp = hay;
+			for(i=0;i<smax;++i) {
+				if(sheadof(needle,sp+i)) {
+					ret = sp+i;
+					break;
+				}
+			}
+		}
+	}
+
+	return ret;
+}
+
+static inline int is_http_get_inline(const char *s, size_t blen) {
+	if(s && blen>=12) {
+		if((s[0]=='G')&&(s[1]=='E')&&(s[2]=='T')&&(s[3]==' ')) {
+			const char *sp=findstr(s+4,blen-4,"HTTP");
+			if(sp) {
+				sp += 4;
+				size_t diff_blen = sp-s;
+				if(diff_blen+4 <= blen) {
+					sp=findstr(sp,blen-diff_blen,"\r\n\r\n");
+					if(sp) {
+						return (int)(sp-s+4);
+					}
+				}
+			}
+
+		}
+	}
+	return 0;
+}
+
+int is_http_get(const char *s, size_t blen) {
+	return is_http_get_inline(s, blen);
+}
+
 int stun_get_message_len_str(u08bits *buf, size_t blen, int padding, size_t *app_len) {
 	if (buf && blen) {
 		/* STUN request/response ? */
@@ -489,9 +606,18 @@ int stun_get_message_len_str(u08bits *buf, size_t blen, int padding, size_t *app
 						}
 					}
 				}
-				return -1;
 			}
 		}
+
+		//HTTP request ?
+		{
+			int http_len = is_http_get_inline(((char*)buf), blen);
+			if((http_len>0) && ((size_t)http_len<=blen)) {
+				*app_len = (size_t)http_len;
+				return http_len;
+			}
+		}
+
 		/* STUN channel ? */
 		if(blen>=4) {
 			u16bits chn=nswap16(((const u16bits*)(buf))[0]);
@@ -510,6 +636,7 @@ int stun_get_message_len_str(u08bits *buf, size_t blen, int padding, size_t *app
 				}
 			}
 		}
+
 	}
 
 	return -1;
@@ -518,7 +645,7 @@ int stun_get_message_len_str(u08bits *buf, size_t blen, int padding, size_t *app
 ////////// ALLOCATE ///////////////////////////////////
 
 int stun_set_allocate_request_str(u08bits* buf, size_t *len, u32bits lifetime, int address_family,
-				u08bits transport) {
+				u08bits transport, int mobile) {
 
   stun_init_request_str(STUN_METHOD_ALLOCATE, buf, len);
 
@@ -539,6 +666,10 @@ int stun_set_allocate_request_str(u08bits* buf, size_t *len, u32bits lifetime, i
     if(stun_attr_add_str(buf,len,STUN_ATTRIBUTE_LIFETIME,(u08bits*)(&field),sizeof(field))<0) return -1;
   }
 
+  if(mobile) {
+	  if(stun_attr_add_str(buf,len,STUN_ATTRIBUTE_MOBILITY_TICKET,(const u08bits*)"",0)<0) return -1;
+  }
+
   //ADRESS-FAMILY
   switch (address_family) {
   case STUN_ATTRIBUTE_REQUESTED_ADDRESS_FAMILY_VALUE_IPV4:
@@ -565,7 +696,7 @@ int stun_set_allocate_request_str(u08bits* buf, size_t *len, u32bits lifetime, i
 int stun_set_allocate_response_str(u08bits* buf, size_t *len, stun_tid* tid, 
 				   const ioa_addr *relayed_addr, const ioa_addr *reflexive_addr,
 				   u32bits lifetime, int error_code, const u08bits *reason,
-				   u64bits reservation_token) {
+				   u64bits reservation_token, char* mobile_id) {
 
   if(!error_code) {
 
@@ -586,10 +717,16 @@ int stun_set_allocate_response_str(u08bits* buf, size_t *len, stun_tid* tid,
 
     {
       if(lifetime<1) lifetime=STUN_DEFAULT_ALLOCATE_LIFETIME;
+      else if(lifetime>STUN_MAX_ALLOCATE_LIFETIME) lifetime = STUN_MAX_ALLOCATE_LIFETIME;
+
       u32bits field=nswap32(lifetime);
       if(stun_attr_add_str(buf,len,STUN_ATTRIBUTE_LIFETIME,(u08bits*)(&field),sizeof(field))<0) return -1;
     }
 
+    if(mobile_id && *mobile_id) {
+	    if(stun_attr_add_str(buf,len,STUN_ATTRIBUTE_MOBILITY_TICKET,(u08bits*)mobile_id,strlen(mobile_id))<0) return -1;
+    }
+
   } else {
     stun_init_error_response_str(STUN_METHOD_ALLOCATE, buf, len, error_code, reason, tid);
   }
@@ -603,7 +740,7 @@ u16bits stun_set_channel_bind_request_str(u08bits* buf, size_t *len,
 					   const ioa_addr* peer_addr, u16bits channel_number) {
 
   if(!STUN_VALID_CHANNEL(channel_number)) {
-    channel_number = 0x4000 + ((u16bits)(((u32bits)random())%(0x7FFF-0x4000+1)));
+    channel_number = 0x4000 + ((u16bits)(((u32bits)turn_random())%(0x7FFF-0x4000+1)));
   }
   
   stun_init_request_str(STUN_METHOD_CHANNEL_BIND, buf, len);
@@ -721,9 +858,8 @@ static void stun_tid_from_string(const u08bits* s, stun_tid* id) {
 }
 
 void stun_tid_from_message_str(const u08bits* buf, size_t len, stun_tid* id) {
-  if(stun_is_command_message_str(buf,len)) {
-    stun_tid_from_string(buf+8, id);
-  }
+	UNUSED_ARG(len);
+	stun_tid_from_string(buf+8, id);
 }
 
 void stun_tid_message_cpy(u08bits* buf, const stun_tid* id) {
@@ -735,9 +871,7 @@ void stun_tid_message_cpy(u08bits* buf, const stun_tid* id) {
 void stun_tid_generate(stun_tid* id) {
   if(id) {
     u32bits *w=(u32bits*)(id->tsx_id);
-    w[0]=(u32bits)random();
-    w[1]=(u32bits)random();
-    w[2]=(u32bits)random();
+    turn_random32_size(w,3);
   }
 }
 
@@ -901,7 +1035,7 @@ int stun_attr_add_str(u08bits* buf, size_t *len, u16bits attr, const u08bits* av
   }
   int clen = stun_get_command_message_len_str(buf,*len);
   int newlen = clen + 4 + alen;
-  int newlenrem4=newlen%4;
+  int newlenrem4=newlen & 0x00000003;
   if(newlenrem4) {
     newlen=newlen+(4-newlenrem4);
   }
@@ -982,7 +1116,7 @@ int stun_attr_get_addr_str(const u08bits *buf, size_t len, stun_attr_ref attr, i
 
   map_addr_from_public_to_private(&public_addr, ca);
 
-  if(default_addr && addr_any_no_port(ca)) {
+  if(default_addr && addr_any_no_port(ca) && !addr_any_no_port(default_addr)) {
     int port = addr_get_port(ca);
     addr_cpy(ca,default_addr);
     addr_set_port(ca,port);
@@ -1191,9 +1325,16 @@ int SASLprep(u08bits *s)
 
 //////////////// Message Integrity ////////////////////////////
 
+size_t get_hmackey_size(SHATYPE shatype)
+{
+	if(shatype == SHATYPE_SHA256)
+		return 32;
+	return 16;
+}
+
 void print_bin_func(const char *name, size_t len, const void *s, const char *func)
 {
-	printf("%s:%s:len=%d:[",func,name,(int)len);
+	printf("<%s>:<%s>:len=%d:[",func,name,(int)len);
 	size_t i;
 	for(i=0;i<len;i++) {
 		printf("%02x",(int)((const u08bits*)s)[i]);
@@ -1201,29 +1342,39 @@ void print_bin_func(const char *name, size_t len, const void *s, const char *fun
 	printf("]\n");
 }
 
-int stun_attr_add_integrity_str(turn_credential_type ct, u08bits *buf, size_t *len, hmackey_t key, st_password_t pwd)
+int stun_attr_add_integrity_str(turn_credential_type ct, u08bits *buf, size_t *len, hmackey_t key, st_password_t pwd, SHATYPE shatype)
 {
-	u08bits hmac[20];
+	u08bits hmac[MAXSHASIZE];
 
-	if(stun_attr_add_str(buf, len, STUN_ATTRIBUTE_MESSAGE_INTEGRITY, hmac, sizeof(hmac))<0)
+	unsigned int shasize;
+
+	switch(shatype) {
+	case SHATYPE_SHA256:
+		shasize = SHA256SIZEBYTES;
+		break;
+	default:
+		shasize = SHA1SIZEBYTES;
+	};
+
+	if(stun_attr_add_str(buf, len, STUN_ATTRIBUTE_MESSAGE_INTEGRITY, hmac, shasize)<0)
 		return -1;
 
 	if(ct == TURN_CREDENTIALS_SHORT_TERM) {
-		if(stun_calculate_hmac(buf, *len-4-sizeof(hmac), pwd, strlen((char*)pwd), buf+*len-sizeof(hmac))<0)
+		if(stun_calculate_hmac(buf, *len-4-shasize, pwd, strlen((char*)pwd), buf+*len-shasize, &shasize, shatype)<0)
 				return -1;
 	} else {
-		if(stun_calculate_hmac(buf, *len-4-sizeof(hmac), key, sizeof(hmackey_t), buf+*len-sizeof(hmac))<0)
+		if(stun_calculate_hmac(buf, *len-4-shasize, key, get_hmackey_size(shatype), buf+*len-shasize, &shasize, shatype)<0)
 			return -1;
 	}
 
 	return 0;
 }
 
-int stun_attr_add_integrity_by_user_str(u08bits *buf, size_t *len, u08bits *uname, u08bits *realm, u08bits *upwd, u08bits *nonce)
+int stun_attr_add_integrity_by_user_str(u08bits *buf, size_t *len, u08bits *uname, u08bits *realm, u08bits *upwd, u08bits *nonce, SHATYPE shatype)
 {
 	hmackey_t key;
 
-	if(stun_produce_integrity_key_str(uname, realm, upwd, key)<0)
+	if(stun_produce_integrity_key_str(uname, realm, upwd, key, shatype)<0)
 		return -1;
 
 	if(stun_attr_add_str(buf, len, STUN_ATTRIBUTE_USERNAME, uname, strlen((s08bits*)uname))<0)
@@ -1236,16 +1387,16 @@ int stun_attr_add_integrity_by_user_str(u08bits *buf, size_t *len, u08bits *unam
 			return -1;
 
 	st_password_t p;
-	return stun_attr_add_integrity_str(TURN_CREDENTIALS_LONG_TERM, buf, len, key, p);
+	return stun_attr_add_integrity_str(TURN_CREDENTIALS_LONG_TERM, buf, len, key, p, shatype);
 }
 
-int stun_attr_add_integrity_by_user_short_term_str(u08bits *buf, size_t *len, u08bits *uname, st_password_t pwd)
+int stun_attr_add_integrity_by_user_short_term_str(u08bits *buf, size_t *len, u08bits *uname, st_password_t pwd, SHATYPE shatype)
 {
 	if(stun_attr_add_str(buf, len, STUN_ATTRIBUTE_USERNAME, uname, strlen((s08bits*)uname))<0)
 			return -1;
 
 	hmackey_t key;
-	return stun_attr_add_integrity_str(TURN_CREDENTIALS_SHORT_TERM, buf, len, key, pwd);
+	return stun_attr_add_integrity_str(TURN_CREDENTIALS_SHORT_TERM, buf, len, key, pwd, shatype);
 }
 
 void print_hmac(const char *name, const void *s, size_t len)
@@ -1261,21 +1412,42 @@ void print_hmac(const char *name, const void *s, size_t len)
 /*
  * Return -1 if failure, 0 if the integrity is not correct, 1 if OK
  */
-int stun_check_message_integrity_by_key_str(turn_credential_type ct, u08bits *buf, size_t len, hmackey_t key, st_password_t pwd)
+int stun_check_message_integrity_by_key_str(turn_credential_type ct, u08bits *buf, size_t len, hmackey_t key, st_password_t pwd, SHATYPE shatype, int *too_weak)
 {
 	int res = 0;
-	u08bits new_hmac[20];
+	u08bits new_hmac[MAXSHASIZE];
+	unsigned int shasize;
 	const u08bits *old_hmac = NULL;
 
 	stun_attr_ref sar = stun_attr_get_first_by_type_str(buf, len, STUN_ATTRIBUTE_MESSAGE_INTEGRITY);
 	if (!sar)
 		return -1;
 
+	int sarlen = stun_attr_get_len(sar);
+
+	switch(sarlen) {
+	case SHA256SIZEBYTES:
+		shasize = SHA256SIZEBYTES;
+		if(shatype != SHATYPE_SHA256)
+			return -1;
+		break;
+	case SHA1SIZEBYTES:
+		shasize = SHA1SIZEBYTES;
+		if(shatype != SHATYPE_SHA1) {
+			if(too_weak)
+				*too_weak = 1;
+			return -1;
+		}
+		break;
+	default:
+		return -1;
+	};
+
 	int orig_len = stun_get_command_message_len_str(buf, len);
 	if (orig_len < 0)
 		return -1;
 
-	int new_len = ((const u08bits*) sar - buf) + 4 + 20;
+	int new_len = ((const u08bits*) sar - buf) + 4 + shasize;
 	if (new_len > orig_len)
 		return -1;
 
@@ -1283,23 +1455,20 @@ int stun_check_message_integrity_by_key_str(turn_credential_type ct, u08bits *bu
 		return -1;
 
 	if(ct == TURN_CREDENTIALS_SHORT_TERM) {
-		res = stun_calculate_hmac(buf, (size_t) new_len - 4 - 20, pwd, strlen((char*)pwd), new_hmac);
+		res = stun_calculate_hmac(buf, (size_t) new_len - 4 - shasize, pwd, strlen((char*)pwd), new_hmac, &shasize, shatype);
 	} else {
-		res = stun_calculate_hmac(buf, (size_t) new_len - 4 - 20, key, sizeof(hmackey_t), new_hmac);
+		res = stun_calculate_hmac(buf, (size_t) new_len - 4 - shasize, key, get_hmackey_size(shatype), new_hmac, &shasize, shatype);
 	}
 
 	stun_set_command_message_len_str(buf, orig_len);
 	if(res<0)
 		return -1;
 
-	if(stun_attr_get_len(sar) != 20)
-		return -1;
-
 	old_hmac = stun_attr_get_value(sar);
 	if(!old_hmac)
 		return -1;
 
-	if(bcmp(old_hmac,new_hmac,sizeof(new_hmac)))
+	if(bcmp(old_hmac,new_hmac,shasize))
 		return 0;
 
 	return 1;
@@ -1308,17 +1477,17 @@ int stun_check_message_integrity_by_key_str(turn_credential_type ct, u08bits *bu
 /*
  * Return -1 if failure, 0 if the integrity is not correct, 1 if OK
  */
-int stun_check_message_integrity_str(turn_credential_type ct, u08bits *buf, size_t len, u08bits *uname, u08bits *realm, u08bits *upwd)
+int stun_check_message_integrity_str(turn_credential_type ct, u08bits *buf, size_t len, u08bits *uname, u08bits *realm, u08bits *upwd, SHATYPE shatype)
 {
 	hmackey_t key;
 	st_password_t pwd;
 
 	if(ct == TURN_CREDENTIALS_SHORT_TERM)
 		strncpy((char*)pwd,(char*)upwd,sizeof(st_password_t));
-	else if (stun_produce_integrity_key_str(uname, realm, upwd, key) < 0)
+	else if (stun_produce_integrity_key_str(uname, realm, upwd, key, shatype) < 0)
 		return -1;
 
-	return stun_check_message_integrity_by_key_str(ct, buf, len, key, pwd);
+	return stun_check_message_integrity_by_key_str(ct, buf, len, key, pwd, shatype, NULL);
 }
 
 /* RFC 5780 */
