@@ -31,8 +31,8 @@
 #ifndef __IOADEFS__
 #define __IOADEFS__
 
-#define TURN_SERVER_VERSION "1.8.7.0"
-#define TURN_SERVER_VERSION_NAME "Black Dow"
+#define TURN_SERVER_VERSION "3.2.3.1"
+#define TURN_SERVER_VERSION_NAME "Marshal West"
 #define TURN_SOFTWARE "Citrix-"TURN_SERVER_VERSION" '"TURN_SERVER_VERSION_NAME"'"
 
 #if (defined(__unix__) || defined(unix)) && !defined(USG)
@@ -126,6 +126,8 @@ typedef u32bits turn_time_t;
 #endif
 
 #define MAX_STUN_MESSAGE_SIZE (65507)
+#define STUN_BUFFER_SIZE (MAX_STUN_MESSAGE_SIZE)
+#define UDP_STUN_BUFFER_SIZE (1024<<4)
 
 #define NONCE_LENGTH_32BITS (4)
 
@@ -150,13 +152,23 @@ typedef u32bits turn_time_t;
 		}\
 	} while(0)
 
-////////////////////////// UDP /////////////////////////
+////////////////// Security ////////////////////////////
 
-#if !defined(TURN_UDP_SOCKET_CONNECT_BUG)
-#define use_colocated_udp_sockets (1)
-#else
-#define use_colocated_udp_sockets (0)
-#endif
+#define SHA1SIZEBYTES (20)
+#define SHA256SIZEBYTES (32)
+
+#define MAXSHASIZE (128)
+
+enum _SHATYPE {
+	SHATYPE_SHA1 = 0,
+	SHATYPE_SHA256
+};
+
+typedef enum _SHATYPE SHATYPE;
+
+#define shatype_name(sht) ((sht == SHATYPE_SHA1) ? "SHA1" : ((sht == SHATYPE_SHA256) ? "SHA256" : "SHA UNKNOWN"))
+
+#define SHA_TOO_WEAK (426)
 
 ////////////////////////////////////////////////////////
 
