@@ -12,4 +12,10 @@ set turn/user/ninefingers/password "youhavetoberealistic"
 set turn/user/gorst/password "hero"
 
 set turn/secret/1368426581 "logen"
+
+set turn/denied-peer-ip/123456 "172.17.13.133-172.17.14.56"
+set turn/denied-peer-ip/234567 "123::45"
+
+set turn/allowed-peer-ip/345678 "172.17.13.200"
+
 !
